from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict


class TravelRuleVasp(BaseModel):
    model_config = ConfigDict(frozen=True)

    vasp_name: str
    vasp_uuid: str
    depositable: bool
    withdrawable: bool


class TravelRuleVerification(BaseModel):
    model_config = ConfigDict(frozen=True)

    deposit_uuid: str
    verification_result: str
    deposit_state: Literal[
        "PROCESSING",
        "ACCEPTED",
        "CANCELLED",
        "REJECTED",
        "TRAVEL_RULE_SUSPECTED",
        "REFUNDING",
        "REFUNDED",
    ]



from __future__ import annotations

import asyncio
import time

from upbeat.strategies._base import AsyncClientProtocol, SyncClientProtocol
from upbeat.types.order import CancelResult, OrderCreated, OrderDetail


def market_buy_krw(
    client: SyncClientProtocol, market: str, krw_amount: float
) -> OrderCreated:
    return client.orders.create(
        market=market,
        side="bid",
        ord_type="price",
        price=str(int(krw_amount)),
    )


def place_and_wait(
    client: SyncClientProtocol,
    *,
    market: str,
    side: str,
    ord_type: str,
    price: str | None = None,
    volume: str | None = None,
    timeout: float,
    poll_interval: float = 1.0,
) -> OrderDetail:
    order = client.orders.create(
        market=market, side=side, ord_type=ord_type, price=price, volume=volume
    )
    deadline = time.monotonic() + timeout
    while True:
        detail = client.orders.get(uuid=order.uuid)
        if detail.state == "done":
            return detail
        if detail.state == "cancel":
            raise RuntimeError(f"Order {order.uuid} was canceled")
        if time.monotonic() >= deadline:
            raise TimeoutError(f"Order {order.uuid} not filled within {timeout}s")
        time.sleep(poll_interval)


def cancel_all_orders(
    client: SyncClientProtocol, market: str | None = None
) -> CancelResult:
    return client.orders.cancel_open(pairs=market)


async def async_market_buy_krw(
    client: AsyncClientProtocol, market: str, krw_amount: float
) -> OrderCreated:
    return await client.orders.create(
        market=market,
        side="bid",
        ord_type="price",
        price=str(int(krw_amount)),
    )


async def async_place_and_wait(
    client: AsyncClientProtocol,
    *,
    market: str,
    side: str,
    ord_type: str,
    price: str | None = None,
    volume: str | None = None,
    timeout: float,
    poll_interval: float = 1.0,
) -> OrderDetail:
    order = await client.orders.create(
        market=market, side=side, ord_type=ord_type, price=price, volume=volume
    )
    deadline = time.monotonic() + timeout
    while True:
        detail = await client.orders.get(uuid=order.uuid)
        if detail.state == "done":
            return detail
        if detail.state == "cancel":
            raise RuntimeError(f"Order {order.uuid} was canceled")
        if time.monotonic() >= deadline:
            raise TimeoutError(f"Order {order.uuid} not filled within {timeout}s")
        await asyncio.sleep(poll_interval)


async def async_cancel_all_orders(
    client: AsyncClientProtocol, market: str | None = None
) -> CancelResult:
    return await client.orders.cancel_open(pairs=market)



from __future__ import annotations

from typing import Protocol

from pydantic import BaseModel, ConfigDict

from upbeat.api.accounts import AccountsAPI, AsyncAccountsAPI
from upbeat.api.orders import AsyncOrdersAPI, OrdersAPI
from upbeat.api.quotation import AsyncQuotationAPI, QuotationAPI
from upbeat.types.order import OrderCreated


class SyncClientProtocol(Protocol):
    @property
    def accounts(self) -> AccountsAPI: ...

    @property
    def orders(self) -> OrdersAPI: ...

    @property
    def quotation(self) -> QuotationAPI: ...


class AsyncClientProtocol(Protocol):
    @property
    def accounts(self) -> AsyncAccountsAPI: ...

    @property
    def orders(self) -> AsyncOrdersAPI: ...

    @property
    def quotation(self) -> AsyncQuotationAPI: ...


class PortfolioItem(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    balance: float
    locked: float
    current_price: float
    market: str
    estimated_value_krw: float


class RebalanceOrder(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    side: str
    ord_type: str
    amount: float
    order: OrderCreated


class RebalanceResult(BaseModel):
    model_config = ConfigDict(frozen=True)

    orders: list[RebalanceOrder]
    total_value: float


class DCAResult(BaseModel):
    model_config = ConfigDict(frozen=True)

    orders: list[OrderCreated]
    total_spent: float
    splits_completed: int



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat._pandas import candles_to_dataframe
from upbeat.types.quotation import (
    CandleDay,
    CandleMinute,
    CandlePeriod,
    CandleSecond,
    Orderbook,
    OrderbookInstrument,
    Ticker,
    Trade,
)

if TYPE_CHECKING:
    import pandas as pd


def _filter_params(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}


class QuotationAPI(_SyncAPIResource):
    def get_tickers(self, markets: str) -> list[Ticker]:
        params = {"markets": markets}
        response = self._transport.request("GET", "/v1/ticker", params=params)
        return [Ticker.model_validate(item) for item in response.data]

    def get_tickers_by_quote(self, quote_currencies: str) -> list[Ticker]:
        params = {"quote_currencies": quote_currencies}
        response = self._transport.request("GET", "/v1/ticker/all", params=params)
        return [Ticker.model_validate(item) for item in response.data]

    def get_candles_minutes(
        self,
        *,
        market: str,
        unit: Literal[1, 3, 5, 10, 15, 30, 60, 240],
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandleMinute]:
        params = _filter_params(market=market, to=to, count=count)
        response = self._transport.request(
            "GET", f"/v1/candles/minutes/{unit}", params=params
        )
        return [CandleMinute.model_validate(item) for item in response.data]

    def get_candles_seconds(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandleSecond]:
        params = _filter_params(market=market, to=to, count=count)
        response = self._transport.request("GET", "/v1/candles/seconds", params=params)
        return [CandleSecond.model_validate(item) for item in response.data]

    def get_candles_days(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
        converting_price_unit: str | None = None,
    ) -> list[CandleDay]:
        params = _filter_params(
            market=market,
            to=to,
            count=count,
            converting_price_unit=converting_price_unit,
        )
        response = self._transport.request("GET", "/v1/candles/days", params=params)
        return [CandleDay.model_validate(item) for item in response.data]

    def get_candles_weeks(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = self._transport.request("GET", "/v1/candles/weeks", params=params)
        return [CandlePeriod.model_validate(item) for item in response.data]

    def get_candles_months(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = self._transport.request("GET", "/v1/candles/months", params=params)
        return [CandlePeriod.model_validate(item) for item in response.data]

    def get_candles_years(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = self._transport.request("GET", "/v1/candles/years", params=params)
        return [CandlePeriod.model_validate(item) for item in response.data]

    def get_candles_minutes_df(
        self,
        *,
        market: str,
        unit: Literal[1, 3, 5, 10, 15, 30, 60, 240],
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_minutes(market=market, unit=unit, to=to, count=count)
        )

    def get_candles_seconds_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_seconds(market=market, to=to, count=count)
        )

    def get_candles_days_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
        converting_price_unit: str | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_days(
                market=market,
                to=to,
                count=count,
                converting_price_unit=converting_price_unit,
            )
        )

    def get_candles_weeks_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_weeks(market=market, to=to, count=count)
        )

    def get_candles_months_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_months(market=market, to=to, count=count)
        )

    def get_candles_years_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            self.get_candles_years(market=market, to=to, count=count)
        )

    def get_orderbooks(
        self,
        markets: str,
        *,
        level: str | None = None,
        count: int | None = None,
    ) -> list[Orderbook]:
        params = _filter_params(markets=markets, level=level, count=count)
        response = self._transport.request("GET", "/v1/orderbook", params=params)
        return [Orderbook.model_validate(item) for item in response.data]

    def get_orderbook_instruments(self, markets: str) -> list[OrderbookInstrument]:
        params = {"markets": markets}
        response = self._transport.request(
            "GET", "/v1/orderbook/instruments", params=params
        )
        return [OrderbookInstrument.model_validate(item) for item in response.data]

    def get_trades(
        self,
        market: str,
        *,
        to: str | None = None,
        count: int | None = None,
        cursor: str | None = None,
        days_ago: int | None = None,
    ) -> list[Trade]:
        params = _filter_params(
            market=market, to=to, count=count, cursor=cursor, days_ago=days_ago
        )
        response = self._transport.request("GET", "/v1/trades/ticks", params=params)
        return [Trade.model_validate(item) for item in response.data]


class AsyncQuotationAPI(_AsyncAPIResource):
    async def get_tickers(self, markets: str) -> list[Ticker]:
        params = {"markets": markets}
        response = await self._transport.request("GET", "/v1/ticker", params=params)
        return [Ticker.model_validate(item) for item in response.data]

    async def get_tickers_by_quote(self, quote_currencies: str) -> list[Ticker]:
        params = {"quote_currencies": quote_currencies}
        response = await self._transport.request("GET", "/v1/ticker/all", params=params)
        return [Ticker.model_validate(item) for item in response.data]

    async def get_candles_minutes(
        self,
        *,
        market: str,
        unit: Literal[1, 3, 5, 10, 15, 30, 60, 240],
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandleMinute]:
        params = _filter_params(market=market, to=to, count=count)
        response = await self._transport.request(
            "GET", f"/v1/candles/minutes/{unit}", params=params
        )
        return [CandleMinute.model_validate(item) for item in response.data]

    async def get_candles_seconds(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandleSecond]:
        params = _filter_params(market=market, to=to, count=count)
        response = await self._transport.request(
            "GET", "/v1/candles/seconds", params=params
        )
        return [CandleSecond.model_validate(item) for item in response.data]

    async def get_candles_days(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
        converting_price_unit: str | None = None,
    ) -> list[CandleDay]:
        params = _filter_params(
            market=market,
            to=to,
            count=count,
            converting_price_unit=converting_price_unit,
        )
        response = await self._transport.request(
            "GET", "/v1/candles/days", params=params
        )
        return [CandleDay.model_validate(item) for item in response.data]

    async def get_candles_weeks(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = await self._transport.request(
            "GET", "/v1/candles/weeks", params=params
        )
        return [CandlePeriod.model_validate(item) for item in response.data]

    async def get_candles_months(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = await self._transport.request(
            "GET", "/v1/candles/months", params=params
        )
        return [CandlePeriod.model_validate(item) for item in response.data]

    async def get_candles_years(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> list[CandlePeriod]:
        params = _filter_params(market=market, to=to, count=count)
        response = await self._transport.request(
            "GET", "/v1/candles/years", params=params
        )
        return [CandlePeriod.model_validate(item) for item in response.data]

    async def get_candles_minutes_df(
        self,
        *,
        market: str,
        unit: Literal[1, 3, 5, 10, 15, 30, 60, 240],
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_minutes(
                market=market,
                unit=unit,
                to=to,
                count=count,
            )
        )

    async def get_candles_seconds_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_seconds(market=market, to=to, count=count)
        )

    async def get_candles_days_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
        converting_price_unit: str | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_days(
                market=market,
                to=to,
                count=count,
                converting_price_unit=converting_price_unit,
            )
        )

    async def get_candles_weeks_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_weeks(market=market, to=to, count=count)
        )

    async def get_candles_months_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_months(market=market, to=to, count=count)
        )

    async def get_candles_years_df(
        self,
        *,
        market: str,
        to: str | None = None,
        count: int | None = None,
    ) -> pd.DataFrame:
        return candles_to_dataframe(
            await self.get_candles_years(market=market, to=to, count=count)
        )

    async def get_orderbooks(
        self,
        markets: str,
        *,
        level: str | None = None,
        count: int | None = None,
    ) -> list[Orderbook]:
        params = _filter_params(markets=markets, level=level, count=count)
        response = await self._transport.request("GET", "/v1/orderbook", params=params)
        return [Orderbook.model_validate(item) for item in response.data]

    async def get_orderbook_instruments(
        self, markets: str
    ) -> list[OrderbookInstrument]:
        params = {"markets": markets}
        response = await self._transport.request(
            "GET", "/v1/orderbook/instruments", params=params
        )
        return [OrderbookInstrument.model_validate(item) for item in response.data]

    async def get_trades(
        self,
        market: str,
        *,
        to: str | None = None,
        count: int | None = None,
        cursor: str | None = None,
        days_ago: int | None = None,
    ) -> list[Trade]:
        params = _filter_params(
            market=market, to=to, count=count, cursor=cursor, days_ago=days_ago
        )
        response = await self._transport.request(
            "GET", "/v1/trades/ticks", params=params
        )
        return [Trade.model_validate(item) for item in response.data]



from __future__ import annotations

import asyncio
import time
from collections.abc import Callable
from typing import Literal

from upbeat.strategies._base import AsyncClientProtocol, SyncClientProtocol


def price_alert(
    client: SyncClientProtocol,
    *,
    market: str,
    stop_loss: float | None = None,
    take_profit: float | None = None,
    callback: Callable[[str, Literal["stop_loss", "take_profit"], float], None],
    poll_interval: float = 5.0,
) -> None:
    if stop_loss is None and take_profit is None:
        raise ValueError("At least one of stop_loss or take_profit must be set")

    while True:
        tickers = client.quotation.get_tickers(market)
        price = tickers[0].trade_price

        if stop_loss is not None and price <= stop_loss:
            callback(market, "stop_loss", price)
            return
        if take_profit is not None and price >= take_profit:
            callback(market, "take_profit", price)
            return

        time.sleep(poll_interval)


async def async_price_alert(
    client: AsyncClientProtocol,
    *,
    market: str,
    stop_loss: float | None = None,
    take_profit: float | None = None,
    callback: Callable[[str, Literal["stop_loss", "take_profit"], float], None],
    poll_interval: float = 5.0,
) -> None:
    if stop_loss is None and take_profit is None:
        raise ValueError("At least one of stop_loss or take_profit must be set")

    while True:
        tickers = await client.quotation.get_tickers(market)
        price = tickers[0].trade_price

        if stop_loss is not None and price <= stop_loss:
            callback(market, "stop_loss", price)
            return
        if take_profit is not None and price >= take_profit:
            callback(market, "take_profit", price)
            return

        await asyncio.sleep(poll_interval)



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict


class TickerMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal["ticker"]
    code: str
    opening_price: float
    high_price: float
    low_price: float
    trade_price: float
    prev_closing_price: float
    change: Literal["RISE", "EVEN", "FALL"]
    change_price: float
    signed_change_price: float
    change_rate: float
    signed_change_rate: float
    trade_volume: float
    acc_trade_volume: float
    acc_trade_volume_24h: float
    acc_trade_price: float
    acc_trade_price_24h: float
    trade_date: str
    trade_time: str
    trade_timestamp: int
    ask_bid: Literal["ASK", "BID"]
    acc_ask_volume: float
    acc_bid_volume: float
    highest_52_week_price: float
    highest_52_week_date: str
    lowest_52_week_price: float
    lowest_52_week_date: str
    market_state: Literal["PREVIEW", "ACTIVE", "DELISTED"]
    is_trading_suspended: bool
    delisting_date: str | None = None
    market_warning: Literal["NONE", "CAUTION"]
    timestamp: int
    stream_type: Literal["SNAPSHOT", "REALTIME"]


class TradeMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal["trade"]
    code: str
    trade_price: float
    trade_volume: float
    ask_bid: Literal["ASK", "BID"]
    prev_closing_price: float
    change: Literal["RISE", "EVEN", "FALL"]
    change_price: float
    trade_date: str
    trade_time: str
    trade_timestamp: int
    timestamp: int
    sequential_id: int
    best_ask_price: float
    best_ask_size: float
    best_bid_price: float
    best_bid_size: float
    stream_type: Literal["SNAPSHOT", "REALTIME"]


class OrderbookUnitMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    ask_price: float
    bid_price: float
    ask_size: float
    bid_size: float


class OrderbookMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal["orderbook"]
    code: str
    total_ask_size: float
    total_bid_size: float
    orderbook_units: list[OrderbookUnitMessage]
    timestamp: int
    level: float
    stream_type: Literal["SNAPSHOT", "REALTIME"]


class CandleMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal[
        "candle.1s",
        "candle.1m",
        "candle.3m",
        "candle.5m",
        "candle.10m",
        "candle.15m",
        "candle.30m",
        "candle.60m",
        "candle.240m",
    ]
    code: str
    candle_date_time_utc: str
    candle_date_time_kst: str
    opening_price: float
    high_price: float
    low_price: float
    trade_price: float
    candle_acc_trade_volume: float
    candle_acc_trade_price: float
    timestamp: int
    stream_type: Literal["SNAPSHOT", "REALTIME"]


class MyOrderMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal["myOrder"]
    code: str
    uuid: str
    ask_bid: Literal["ASK", "BID"]
    order_type: Literal["limit", "price", "market", "best"]
    state: Literal["wait", "watch", "trade", "done", "cancel", "prevented"]
    price: float
    avg_price: float
    volume: float
    remaining_volume: float
    executed_volume: float
    trades_count: int
    reserved_fee: float
    remaining_fee: float
    paid_fee: float
    locked: float
    executed_funds: float
    trade_timestamp: int
    order_timestamp: int
    timestamp: int
    stream_type: Literal["SNAPSHOT", "REALTIME"]
    trade_uuid: str | None = None
    time_in_force: Literal["ioc", "fok", "post_only"] | None = None
    trade_fee: float | None = None
    is_maker: bool | None = None
    identifier: str | None = None
    smp_type: Literal["reduce", "cancel_maker", "cancel_taker"] | None = None
    prevented_volume: float | None = None
    prevented_locked: float | None = None


class MyAssetItem(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    balance: float
    locked: float


class MyAssetMessage(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: Literal["myAsset"]
    asset_uuid: str
    assets: list[MyAssetItem]
    asset_timestamp: int
    timestamp: int
    stream_type: Literal["REALTIME"]


# ── parse_message ────────────────────────────────────────────────────────

type WebSocketMessage = (
    TickerMessage
    | TradeMessage
    | OrderbookMessage
    | CandleMessage
    | MyOrderMessage
    | MyAssetMessage
)

_TYPE_TO_MODEL: dict[str, type[WebSocketMessage]] = {
    "ticker": TickerMessage,
    "trade": TradeMessage,
    "orderbook": OrderbookMessage,
    "myOrder": MyOrderMessage,
    "myAsset": MyAssetMessage,
}


def parse_message(data: dict[str, object]) -> WebSocketMessage:
    msg_type = data.get("type", "")
    if isinstance(msg_type, str) and msg_type.startswith("candle."):
        return CandleMessage.model_validate(data)
    model = _TYPE_TO_MODEL.get(msg_type)  # type: ignore[arg-type]
    if model is None:
        raise ValueError(f"Unknown WebSocket message type: {msg_type}")
    return model.model_validate(data)



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from upbeat.types.quotation import CandleBase

if TYPE_CHECKING:
    import pandas as pd

_COLUMNS = [
    "market",
    "open",
    "high",
    "low",
    "close",
    "volume",
    "timestamp",
]


def candles_to_dataframe(candles: Sequence[CandleBase]) -> pd.DataFrame:
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "pandas is required for DataFrame support. "
            "Install it with: pip install 'upbeat[pandas]'"
        ) from None

    if not candles:
        df = pd.DataFrame(columns=_COLUMNS)
        df.index = pd.DatetimeIndex([], name="datetime", tz="UTC")
        return df

    rows = [
        {
            "market": c.market,
            "open": c.opening_price,
            "high": c.high_price,
            "low": c.low_price,
            "close": c.trade_price,
            "volume": c.candle_acc_trade_volume,
            "timestamp": c.timestamp,
        }
        for c in candles
    ]
    df = pd.DataFrame(rows)
    df.index = pd.DatetimeIndex(
        pd.to_datetime([c.candle_date_time_utc for c in candles], utc=True),
        name="datetime",
    )
    return df



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class Account(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    balance: str
    locked: str
    avg_buy_price: str
    avg_buy_price_modified: bool
    unit_currency: str



from __future__ import annotations

from pydantic import BaseModel, ConfigDict

# ── OrderTrade (체결 내역, OrderDetail 내부) ────────────────────────────


class OrderTrade(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    price: str
    volume: str
    funds: str
    trend: str
    created_at: str
    side: str


# ── OrderCreated (POST /v1/orders 응답) ─────────────────────────────────


class OrderCreated(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str
    executed_volume: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    trades_count: int
    time_in_force: str | None = None
    identifier: str | None = None
    smp_type: str | None = None
    prevented_volume: str
    prevented_locked: str


# ── OrderDetail (GET /v1/order 응답) ────────────────────────────────────


class OrderDetail(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str | None = None
    executed_volume: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    time_in_force: str | None = None
    smp_type: str | None = None
    prevented_volume: str | None = None
    prevented_locked: str
    identifier: str | None = None
    trades_count: int
    trades: list[OrderTrade]


# ── OrderOpen (GET /v1/orders/open 응답) ────────────────────────────────


class OrderOpen(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str
    executed_volume: str
    executed_funds: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    time_in_force: str | None = None
    identifier: str | None = None
    smp_type: str | None = None
    prevented_volume: str
    prevented_locked: str
    trades_count: int


# ── OrderClosed (GET /v1/orders/closed 응답) ────────────────────────────


class OrderClosed(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str
    state: str
    created_at: str
    volume: str
    remaining_volume: str
    executed_volume: str
    executed_funds: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    time_in_force: str | None = None
    identifier: str | None = None
    smp_type: str | None = None
    prevented_volume: str
    prevented_locked: str
    trades_count: int


# ── OrderByIds (GET /v1/orders/uuids 응답) ──────────────────────────────


class OrderByIds(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str | None = None
    executed_volume: str
    executed_funds: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    time_in_force: str | None = None
    identifier: str | None = None
    smp_type: str | None = None
    prevented_volume: str | None = None
    prevented_locked: str
    trades_count: int


# ── OrderCanceled (DELETE /v1/order 응답) ───────────────────────────────


class OrderCanceled(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str
    executed_volume: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    time_in_force: str | None = None
    identifier: str | None = None
    smp_type: str | None = None
    prevented_volume: str
    prevented_locked: str
    trades_count: int


# ── CancelResult (DELETE /v1/orders/open, /uuids 응답) ─────────────────


class CancelResultOrder(BaseModel):
    model_config = ConfigDict(frozen=True)

    uuid: str
    market: str
    identifier: str | None = None


class CancelResultGroup(BaseModel):
    model_config = ConfigDict(frozen=True)

    count: int
    orders: list[CancelResultOrder]


class CancelResult(BaseModel):
    model_config = ConfigDict(frozen=True)

    success: CancelResultGroup
    failed: CancelResultGroup


# ── CancelAndNewOrderResponse (POST /v1/orders/cancel_and_new 응답) ─────


class CancelAndNewOrderResponse(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    uuid: str
    identifier: str | None = None
    side: str
    ord_type: str
    price: str | None = None
    state: str
    created_at: str
    volume: str | None = None
    remaining_volume: str
    executed_volume: str
    reserved_fee: str
    remaining_fee: str
    paid_fee: str
    locked: str
    prevented_volume: str
    prevented_locked: str
    smp_type: str | None = None
    trades_count: int
    time_in_force: str | None = None
    new_order_uuid: str
    new_order_identifier: str | None = None


# ── OrderChance (GET /v1/orders/chance 응답) ────────────────────────────


class OrderChanceMarketBidAsk(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    min_total: str


class OrderChanceMarket(BaseModel):
    model_config = ConfigDict(frozen=True)

    id: str
    name: str
    order_types: list[str] | None = None
    order_sides: list[str] | None = None
    bid_types: list[str] | None = None
    ask_types: list[str] | None = None
    bid: OrderChanceMarketBidAsk | None = None
    ask: OrderChanceMarketBidAsk | None = None
    max_total: str | None = None
    state: str | None = None


class OrderChanceAccount(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    balance: str
    locked: str
    avg_buy_price: str
    avg_buy_price_modified: bool
    unit_currency: str


class OrderChance(BaseModel):
    model_config = ConfigDict(frozen=True)

    bid_fee: str
    ask_fee: str
    maker_bid_fee: str
    maker_ask_fee: str
    market: OrderChanceMarket
    bid_account: OrderChanceAccount
    ask_account: OrderChanceAccount



from upbeat.ws._client import AsyncUpbeatWebSocket, WebSocketConnection
from upbeat.ws._payload import generate_multi_payload, generate_payload
from upbeat.ws.types import (
    CandleMessage,
    MyAssetItem,
    MyAssetMessage,
    MyOrderMessage,
    OrderbookMessage,
    OrderbookUnitMessage,
    TickerMessage,
    TradeMessage,
    WebSocketMessage,
    parse_message,
)

__all__ = [
    "AsyncUpbeatWebSocket",
    "CandleMessage",
    "MyAssetItem",
    "MyAssetMessage",
    "MyOrderMessage",
    "OrderbookMessage",
    "OrderbookUnitMessage",
    "TickerMessage",
    "TradeMessage",
    "WebSocketConnection",
    "WebSocketMessage",
    "generate_multi_payload",
    "generate_payload",
    "parse_message",
]



from dataclasses import dataclass
from typing import Any

from upbeat._errors import RemainingRequest


@dataclass(frozen=True, slots=True)
class APIResponse:
    data: Any
    remaining_request: RemainingRequest | None
    status_code: int



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict


class Ticker(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    trade_date: str
    trade_time: str
    trade_date_kst: str
    trade_time_kst: str
    trade_timestamp: int
    opening_price: float
    high_price: float
    low_price: float
    trade_price: float
    prev_closing_price: float
    change: Literal["EVEN", "RISE", "FALL"]
    change_price: float
    change_rate: float
    signed_change_price: float
    signed_change_rate: float
    trade_volume: float
    acc_trade_price: float
    acc_trade_price_24h: float
    acc_trade_volume: float
    acc_trade_volume_24h: float
    highest_52_week_price: float
    highest_52_week_date: str
    lowest_52_week_price: float
    lowest_52_week_date: str
    timestamp: int


# ── Candles ──────────────────────────────────────────────────────────────


class CandleBase(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    candle_date_time_utc: str
    candle_date_time_kst: str
    opening_price: float
    high_price: float
    low_price: float
    trade_price: float
    timestamp: int
    candle_acc_trade_price: float
    candle_acc_trade_volume: float


class CandleMinute(CandleBase):
    unit: int


class CandleSecond(CandleBase):
    pass


class CandleDay(CandleBase):
    prev_closing_price: float
    change_price: float
    change_rate: float
    converted_trade_price: float | None = None


class CandlePeriod(CandleBase):
    first_day_of_period: str


# ── Orderbook ────────────────────────────────────────────────────────────


class OrderbookUnit(BaseModel):
    model_config = ConfigDict(frozen=True)

    ask_price: float
    bid_price: float
    ask_size: float
    bid_size: float


class Orderbook(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    timestamp: int
    total_ask_size: float
    total_bid_size: float
    orderbook_units: list[OrderbookUnit]
    level: float


class OrderbookInstrument(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    quote_currency: str
    tick_size: str
    supported_levels: list[str]


# ── Trade ────────────────────────────────────────────────────────────────


class Trade(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    trade_date_utc: str
    trade_time_utc: str
    timestamp: int
    trade_price: float
    trade_volume: float
    prev_closing_price: float
    change_price: float
    ask_bid: Literal["ASK", "BID"]
    sequential_id: int



from upbeat.api.accounts import AccountsAPI, AsyncAccountsAPI
from upbeat.api.deposits import AsyncDepositsAPI, DepositsAPI
from upbeat.api.markets import AsyncMarketsAPI, MarketsAPI
from upbeat.api.orders import AsyncOrdersAPI, OrdersAPI
from upbeat.api.quotation import AsyncQuotationAPI, QuotationAPI
from upbeat.api.travel_rule import AsyncTravelRuleAPI, TravelRuleAPI
from upbeat.api.withdrawals import AsyncWithdrawalsAPI, WithdrawalsAPI

__all__ = [
    "AccountsAPI",
    "AsyncAccountsAPI",
    "AsyncDepositsAPI",
    "AsyncMarketsAPI",
    "AsyncOrdersAPI",
    "AsyncQuotationAPI",
    "AsyncTravelRuleAPI",
    "AsyncWithdrawalsAPI",
    "DepositsAPI",
    "MarketsAPI",
    "OrdersAPI",
    "QuotationAPI",
    "TravelRuleAPI",
    "WithdrawalsAPI",
]



from __future__ import annotations

import uuid

_PUBLIC_CHANNELS = {"ticker", "trade", "orderbook"}
_CANDLE_PREFIXES = "candle."


def generate_payload(
    *,
    type: str,
    codes: list[str] | None = None,
    is_only_snapshot: bool = False,
    is_only_realtime: bool = False,
    level: float | None = None,
    ticket: str | None = None,
    format: str = "DEFAULT",
) -> list[dict[str, object]]:
    ticket_value = ticket or str(uuid.uuid4())

    is_public = type in _PUBLIC_CHANNELS or type.startswith(_CANDLE_PREFIXES)

    if type == "myAsset" and codes is not None:
        raise ValueError("myAsset channel does not accept codes parameter")

    if type in ("myOrder", "myAsset") and (is_only_snapshot or is_only_realtime):
        raise ValueError(
            f"{type} channel does not support is_only_snapshot/is_only_realtime"
        )

    if is_public and not codes:
        raise ValueError(f"{type} channel requires codes parameter")

    type_obj: dict[str, object] = {"type": type}

    if codes is not None:
        type_obj["codes"] = codes

    if is_only_snapshot:
        type_obj["is_only_snapshot"] = True
    if is_only_realtime:
        type_obj["is_only_realtime"] = True

    if level is not None and type == "orderbook":
        type_obj["level"] = level

    return [
        {"ticket": ticket_value},
        type_obj,
        {"format": format},
    ]


def generate_multi_payload(
    subscriptions: list[dict[str, object]],
    *,
    ticket: str | None = None,
    format: str = "DEFAULT",
) -> list[dict[str, object]]:
    ticket_value = ticket or str(uuid.uuid4())
    result: list[dict[str, object]] = [{"ticket": ticket_value}]
    result.extend(subscriptions)
    result.append({"format": format})
    return result



from __future__ import annotations

import asyncio
import time

from upbeat.strategies._base import AsyncClientProtocol, DCAResult, SyncClientProtocol
from upbeat.strategies.order_helpers import async_market_buy_krw, market_buy_krw
from upbeat.types.order import OrderCreated


def dca(
    client: SyncClientProtocol,
    *,
    market: str,
    total_amount: float,
    splits: int,
    interval_seconds: float,
) -> DCAResult:
    amount_per_split = total_amount / splits
    orders: list[OrderCreated] = []

    for i in range(splits):
        try:
            order = market_buy_krw(client, market, amount_per_split)
            orders.append(order)
        except Exception:
            pass

        if i < splits - 1:
            time.sleep(interval_seconds)

    return DCAResult(
        orders=orders,
        total_spent=amount_per_split * len(orders),
        splits_completed=len(orders),
    )


async def async_dca(
    client: AsyncClientProtocol,
    *,
    market: str,
    total_amount: float,
    splits: int,
    interval_seconds: float,
) -> DCAResult:
    amount_per_split = total_amount / splits
    orders: list[OrderCreated] = []

    for i in range(splits):
        try:
            order = await async_market_buy_krw(client, market, amount_per_split)
            orders.append(order)
        except Exception:
            pass

        if i < splits - 1:
            await asyncio.sleep(interval_seconds)

    return DCAResult(
        orders=orders,
        total_spent=amount_per_split * len(orders),
        splits_completed=len(orders),
    )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Protocol, runtime_checkable


@dataclass(frozen=True, slots=True)
class RequestInfo:
    method: str
    url: str
    headers: dict[str, str]


@dataclass(frozen=True, slots=True)
class ResponseInfo:
    status_code: int
    url: str
    headers: dict[str, str]
    elapsed_ms: float


@dataclass(frozen=True, slots=True)
class RetryInfo:
    attempt: int
    max_retries: int
    delay_seconds: float
    reason: str
    method: str
    url: str


@dataclass(frozen=True, slots=True)
class ErrorInfo:
    error: Exception
    method: str
    url: str


@dataclass(frozen=True, slots=True)
class WebSocketEventInfo:
    event: str
    url: str
    message: str | None = None
    error: Exception | None = None


@runtime_checkable
class Logger(Protocol):
    def on_request(self, info: RequestInfo) -> None: ...
    def on_response(self, info: ResponseInfo) -> None: ...
    def on_retry(self, info: RetryInfo) -> None: ...
    def on_error(self, info: ErrorInfo) -> None: ...
    def on_ws_event(self, info: WebSocketEventInfo) -> None: ...


class DefaultLogger:
    def __init__(self) -> None:
        self._logger = logging.getLogger("upbeat")
        self._logger.addHandler(logging.NullHandler())

    def on_request(self, info: RequestInfo) -> None:
        self._logger.debug("Request: %s %s", info.method, info.url)

    def on_response(self, info: ResponseInfo) -> None:
        self._logger.debug(
            "Response: %s %s (%.1fms)", info.status_code, info.url, info.elapsed_ms
        )

    def on_retry(self, info: RetryInfo) -> None:
        self._logger.warning(
            "Retry %s/%s for %s %s in %.1fs: %s",
            info.attempt,
            info.max_retries,
            info.method,
            info.url,
            info.delay_seconds,
            info.reason,
        )

    def on_error(self, info: ErrorInfo) -> None:
        self._logger.error("Error on %s %s: %s", info.method, info.url, info.error)

    def on_ws_event(self, info: WebSocketEventInfo) -> None:
        if info.error is not None:
            self._logger.warning(
                "WebSocket %s %s: %s", info.event, info.url, info.error
            )
        else:
            self._logger.debug("WebSocket %s %s", info.event, info.url)


class NullLogger:
    def on_request(self, info: RequestInfo) -> None:
        pass

    def on_response(self, info: ResponseInfo) -> None:
        pass

    def on_retry(self, info: RetryInfo) -> None:
        pass

    def on_error(self, info: ErrorInfo) -> None:
        pass

    def on_ws_event(self, info: WebSocketEventInfo) -> None:
        pass



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class Deposit(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: str
    uuid: str
    currency: str
    net_type: str | None = None
    txid: str
    state: str
    created_at: str
    done_at: str | None = None
    amount: str
    fee: str
    transaction_type: str


class DepositAddress(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    net_type: str | None = None
    deposit_address: str | None = None
    secondary_address: str | None = None


class DepositAddressCreated(BaseModel):
    model_config = ConfigDict(frozen=True)

    success: bool
    message: str


class DepositChanceCoin(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    net_type: str | None = None
    is_deposit_possible: bool
    deposit_impossible_reason: str
    minimum_deposit_amount: str
    minimum_deposit_confirmations: int
    decimal_precision: int



from __future__ import annotations

from typing import TYPE_CHECKING

from upbeat._client import Upbeat
from upbeat.types.market import TradingPair
from upbeat.types.quotation import (
    CandleDay,
    CandleMinute,
    CandlePeriod,
    CandleSecond,
    Orderbook,
    Ticker,
    Trade,
)

if TYPE_CHECKING:
    import pandas as pd

_MINUTE_UNITS: dict[str, int] = {
    "1m": 1,
    "3m": 3,
    "5m": 5,
    "10m": 10,
    "15m": 15,
    "30m": 30,
    "60m": 60,
    "240m": 240,
}


def get_ticker(market: str) -> Ticker:
    with Upbeat() as client:
        return client.quotation.get_tickers(market)[0]


def get_tickers(markets: str) -> list[Ticker]:
    with Upbeat() as client:
        return client.quotation.get_tickers(markets)


def get_candles(
    market: str,
    *,
    interval: str,
    to: str | None = None,
    count: int | None = None,
) -> list[CandleMinute] | list[CandleSecond] | list[CandleDay] | list[CandlePeriod]:
    with Upbeat() as client:
        if interval in _MINUTE_UNITS:
            return client.quotation.get_candles_minutes(
                market=market,
                unit=_MINUTE_UNITS[interval],
                to=to,
                count=count,  # type: ignore[arg-type]
            )
        if interval == "1s":
            return client.quotation.get_candles_seconds(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1d":
            return client.quotation.get_candles_days(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1w":
            return client.quotation.get_candles_weeks(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1M":
            return client.quotation.get_candles_months(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1y":
            return client.quotation.get_candles_years(
                market=market,
                to=to,
                count=count,
            )
        raise ValueError(
            f"Invalid interval: {interval!r}. "
            f"Expected one of: 1s, 1m, 3m, 5m, 10m, 15m, 30m, 60m, 240m, 1d, 1w, 1M, 1y"
        )


def get_candles_df(
    market: str,
    *,
    interval: str,
    to: str | None = None,
    count: int | None = None,
) -> pd.DataFrame:
    with Upbeat() as client:
        if interval in _MINUTE_UNITS:
            return client.quotation.get_candles_minutes_df(
                market=market,
                unit=_MINUTE_UNITS[interval],
                to=to,
                count=count,  # type: ignore[arg-type]
            )
        if interval == "1s":
            return client.quotation.get_candles_seconds_df(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1d":
            return client.quotation.get_candles_days_df(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1w":
            return client.quotation.get_candles_weeks_df(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1M":
            return client.quotation.get_candles_months_df(
                market=market,
                to=to,
                count=count,
            )
        if interval == "1y":
            return client.quotation.get_candles_years_df(
                market=market,
                to=to,
                count=count,
            )
        raise ValueError(
            f"Invalid interval: {interval!r}. "
            f"Expected one of: 1s, 1m, 3m, 5m, 10m, 15m, 30m, 60m, 240m, 1d, 1w, 1M, 1y"
        )


def get_orderbook(market: str) -> Orderbook:
    with Upbeat() as client:
        return client.quotation.get_orderbooks(market)[0]


def get_orderbooks(markets: str) -> list[Orderbook]:
    with Upbeat() as client:
        return client.quotation.get_orderbooks(markets)


def get_trades(
    market: str,
    *,
    to: str | None = None,
    count: int | None = None,
    cursor: str | None = None,
    days_ago: int | None = None,
) -> list[Trade]:
    with Upbeat() as client:
        return client.quotation.get_trades(
            market,
            to=to,
            count=count,
            cursor=cursor,
            days_ago=days_ago,
        )


def get_markets(*, is_details: bool = False) -> list[TradingPair]:
    with Upbeat() as client:
        return client.markets.get_trading_pairs(is_details=is_details)



from __future__ import annotations

from typing import Any

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.withdrawal import (
    Withdrawal,
    WithdrawalAddress,
    WithdrawalChance,
    WithdrawalKrw,
)


def _filter_params(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}


class WithdrawalsAPI(_SyncAPIResource):
    def get(
        self,
        *,
        uuid: str | None = None,
        txid: str | None = None,
        currency: str | None = None,
    ) -> Withdrawal:
        params = _filter_params(uuid=uuid, txid=txid, currency=currency)
        response = self._transport.request(
            "GET", "/v1/withdraw", params=params, credentials=self._credentials
        )
        return Withdrawal.model_validate(response.data)

    def list(
        self,
        *,
        currency: str | None = None,
        state: str | None = None,
        uuids: list[str] | None = None,
        txids: list[str] | None = None,
        limit: int | None = None,
        page: int | None = None,
        order_by: str | None = None,
        from_cursor: str | None = None,
        to: str | None = None,
    ) -> list[Withdrawal]:
        params: dict[str, Any] = _filter_params(
            currency=currency,
            state=state,
            limit=limit,
            page=page,
            order_by=order_by,
            to=to,
        )
        if from_cursor is not None:
            params["from"] = from_cursor
        if uuids is not None:
            params["uuids[]"] = uuids
        if txids is not None:
            params["txids[]"] = txids
        response = self._transport.request(
            "GET", "/v1/withdraws", params=params, credentials=self._credentials
        )
        return [Withdrawal.model_validate(item) for item in response.data]

    def create_coin(
        self,
        *,
        currency: str,
        net_type: str,
        amount: str,
        address: str,
        secondary_address: str | None = None,
        transaction_type: str | None = None,
    ) -> Withdrawal:
        json_body = _filter_params(
            currency=currency,
            net_type=net_type,
            amount=amount,
            address=address,
            secondary_address=secondary_address,
            transaction_type=transaction_type,
        )
        response = self._transport.request(
            "POST",
            "/v1/withdraws/coin",
            json_body=json_body,
            credentials=self._credentials,
        )
        return Withdrawal.model_validate(response.data)

    def cancel_coin(self, *, uuid: str) -> Withdrawal:
        params = {"uuid": uuid}
        response = self._transport.request(
            "DELETE",
            "/v1/withdraws/coin",
            params=params,
            credentials=self._credentials,
        )
        return Withdrawal.model_validate(response.data)

    def create_krw(
        self,
        *,
        amount: str,
        two_factor_type: str,
    ) -> WithdrawalKrw:
        json_body = {"amount": amount, "two_factor_type": two_factor_type}
        response = self._transport.request(
            "POST",
            "/v1/withdraws/krw",
            json_body=json_body,
            credentials=self._credentials,
        )
        return WithdrawalKrw.model_validate(response.data)

    def list_coin_addresses(self) -> list[WithdrawalAddress]:
        response = self._transport.request(
            "GET",
            "/v1/withdraws/coin_addresses",
            credentials=self._credentials,
        )
        return [WithdrawalAddress.model_validate(item) for item in response.data]

    def get_chance(
        self,
        *,
        currency: str,
        net_type: str | None = None,
    ) -> WithdrawalChance:
        params = _filter_params(currency=currency, net_type=net_type)
        response = self._transport.request(
            "GET",
            "/v1/withdraws/chance",
            params=params,
            credentials=self._credentials,
        )
        return WithdrawalChance.model_validate(response.data)


class AsyncWithdrawalsAPI(_AsyncAPIResource):
    async def get(
        self,
        *,
        uuid: str | None = None,
        txid: str | None = None,
        currency: str | None = None,
    ) -> Withdrawal:
        params = _filter_params(uuid=uuid, txid=txid, currency=currency)
        response = await self._transport.request(
            "GET", "/v1/withdraw", params=params, credentials=self._credentials
        )
        return Withdrawal.model_validate(response.data)

    async def list(
        self,
        *,
        currency: str | None = None,
        state: str | None = None,
        uuids: list[str] | None = None,
        txids: list[str] | None = None,
        limit: int | None = None,
        page: int | None = None,
        order_by: str | None = None,
        from_cursor: str | None = None,
        to: str | None = None,
    ) -> list[Withdrawal]:
        params: dict[str, Any] = _filter_params(
            currency=currency,
            state=state,
            limit=limit,
            page=page,
            order_by=order_by,
            to=to,
        )
        if from_cursor is not None:
            params["from"] = from_cursor
        if uuids is not None:
            params["uuids[]"] = uuids
        if txids is not None:
            params["txids[]"] = txids
        response = await self._transport.request(
            "GET", "/v1/withdraws", params=params, credentials=self._credentials
        )
        return [Withdrawal.model_validate(item) for item in response.data]

    async def create_coin(
        self,
        *,
        currency: str,
        net_type: str,
        amount: str,
        address: str,
        secondary_address: str | None = None,
        transaction_type: str | None = None,
    ) -> Withdrawal:
        json_body = _filter_params(
            currency=currency,
            net_type=net_type,
            amount=amount,
            address=address,
            secondary_address=secondary_address,
            transaction_type=transaction_type,
        )
        response = await self._transport.request(
            "POST",
            "/v1/withdraws/coin",
            json_body=json_body,
            credentials=self._credentials,
        )
        return Withdrawal.model_validate(response.data)

    async def cancel_coin(self, *, uuid: str) -> Withdrawal:
        params = {"uuid": uuid}
        response = await self._transport.request(
            "DELETE",
            "/v1/withdraws/coin",
            params=params,
            credentials=self._credentials,
        )
        return Withdrawal.model_validate(response.data)

    async def create_krw(
        self,
        *,
        amount: str,
        two_factor_type: str,
    ) -> WithdrawalKrw:
        json_body = {"amount": amount, "two_factor_type": two_factor_type}
        response = await self._transport.request(
            "POST",
            "/v1/withdraws/krw",
            json_body=json_body,
            credentials=self._credentials,
        )
        return WithdrawalKrw.model_validate(response.data)

    async def list_coin_addresses(self) -> list[WithdrawalAddress]:
        response = await self._transport.request(
            "GET",
            "/v1/withdraws/coin_addresses",
            credentials=self._credentials,
        )
        return [WithdrawalAddress.model_validate(item) for item in response.data]

    async def get_chance(
        self,
        *,
        currency: str,
        net_type: str | None = None,
    ) -> WithdrawalChance:
        params = _filter_params(currency=currency, net_type=net_type)
        response = await self._transport.request(
            "GET",
            "/v1/withdraws/chance",
            params=params,
            credentials=self._credentials,
        )
        return WithdrawalChance.model_validate(response.data)



from __future__ import annotations

from upbeat._auth import Credentials
from upbeat._http import AsyncTransport, SyncTransport


class _SyncAPIResource:
    def __init__(
        self, transport: SyncTransport, credentials: Credentials | None
    ) -> None:
        self._transport = transport
        self._credentials = credentials


class _AsyncAPIResource:
    def __init__(
        self, transport: AsyncTransport, credentials: Credentials | None
    ) -> None:
        self._transport = transport
        self._credentials = credentials



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import httpx


@dataclass(frozen=True, slots=True)
class RemainingRequest:
    group: str
    remaining_min: int
    remaining_sec: int

    @classmethod
    def from_header(cls, header_value: str) -> RemainingRequest:
        """Parse ``"group=default; min=1800; sec=29"`` format."""
        parts: dict[str, str] = {}
        for item in header_value.split(";"):
            key, _, value = item.strip().partition("=")
            parts[key.strip()] = value.strip()
        return cls(
            group=parts.get("group", ""),
            remaining_min=int(parts.get("min", 0)),
            remaining_sec=int(parts.get("sec", 0)),
        )


# ── Base ─────────────────────────────────────────────────────────────────


class UpbeatError(Exception):
    message: str

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


# ── API errors ───────────────────────────────────────────────────────────


class APIError(UpbeatError):
    pass


class APIStatusError(APIError):
    status_code: int
    error_name: str | None
    error_message: str | None
    response: httpx.Response
    remaining_request: RemainingRequest | None

    def __init__(
        self,
        message: str,
        *,
        status_code: int,
        error_name: str | None,
        error_message: str | None,
        response: httpx.Response,
        remaining_request: RemainingRequest | None,
    ) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.error_name = error_name
        self.error_message = error_message
        self.response = response
        self.remaining_request = remaining_request

    @classmethod
    def from_response(cls, response: httpx.Response) -> APIStatusError:
        status_code = response.status_code
        error_name: str | None = None
        error_message: str | None = None

        try:
            body = response.json()
            error_obj = body.get("error", {})
            error_name = error_obj.get("name")
            error_message = error_obj.get("message")
        except Exception:
            error_message = response.text

        remaining_request: RemainingRequest | None = None
        remaining_header = response.headers.get("Remaining-Req")
        if remaining_header:
            remaining_request = RemainingRequest.from_header(remaining_header)

        message = error_message or f"HTTP {status_code}"

        kwargs = {
            "message": message,
            "status_code": status_code,
            "error_name": error_name,
            "error_message": error_message,
            "response": response,
            "remaining_request": remaining_request,
        }

        # Pattern-based subclass matching
        if error_name and re.match(r"^insufficient_funds_", error_name):
            return InsufficientFundsError(**kwargs)
        if error_name and re.match(r"^under_min_total_", error_name):
            return MinimumOrderError(**kwargs)

        # Status code mapping
        subclass = _STATUS_CODE_MAP.get(status_code)
        if subclass is not None:
            return subclass(**kwargs)

        if status_code >= 500:
            return InternalServerError(**kwargs)

        return APIStatusError(**kwargs)


class BadRequestError(APIStatusError):
    pass


class AuthenticationError(APIStatusError):
    pass


class PermissionDeniedError(APIStatusError):
    pass


class NotFoundError(APIStatusError):
    pass


class UnprocessableEntityError(APIStatusError):
    pass


class RateLimitError(APIStatusError):
    pass


class InternalServerError(APIStatusError):
    pass


class InsufficientFundsError(BadRequestError):
    pass


class MinimumOrderError(BadRequestError):
    pass


_STATUS_CODE_MAP: dict[int, type[APIStatusError]] = {
    400: BadRequestError,
    401: AuthenticationError,
    403: PermissionDeniedError,
    404: NotFoundError,
    418: RateLimitError,
    422: UnprocessableEntityError,
    429: RateLimitError,
}


# ── Connection errors ────────────────────────────────────────────────────


class APIConnectionError(APIError):
    cause: Exception | None

    def __init__(self, message: str, *, cause: Exception | None = None) -> None:
        super().__init__(message)
        self.cause = cause


class APITimeoutError(APIConnectionError):
    pass


# ── WebSocket errors ─────────────────────────────────────────────────────


class WebSocketError(UpbeatError):
    pass


class WebSocketConnectionError(WebSocketError):
    pass


class WebSocketClosedError(WebSocketError):
    code: int | None
    reason: str | None

    def __init__(
        self,
        message: str,
        *,
        code: int | None = None,
        reason: str | None = None,
    ) -> None:
        super().__init__(message)
        self.code = code
        self.reason = reason



from __future__ import annotations

from typing import Any

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.order import (
    CancelAndNewOrderResponse,
    CancelResult,
    OrderByIds,
    OrderCanceled,
    OrderChance,
    OrderClosed,
    OrderCreated,
    OrderDetail,
    OrderOpen,
)


def _filter_params(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}


class OrdersAPI(_SyncAPIResource):
    def create(
        self,
        *,
        market: str,
        side: str,
        volume: str | None = None,
        price: str | None = None,
        ord_type: str = "limit",
        identifier: str | None = None,
        time_in_force: str | None = None,
        smp_type: str | None = None,
    ) -> OrderCreated:
        json_body = _filter_params(
            market=market,
            side=side,
            volume=volume,
            price=price,
            ord_type=ord_type,
            identifier=identifier,
            time_in_force=time_in_force,
            smp_type=smp_type,
        )
        response = self._transport.request(
            "POST", "/v1/orders", json_body=json_body, credentials=self._credentials
        )
        return OrderCreated.model_validate(response.data)

    def create_test(
        self,
        *,
        market: str,
        side: str,
        volume: str | None = None,
        price: str | None = None,
        ord_type: str = "limit",
        identifier: str | None = None,
        time_in_force: str | None = None,
        smp_type: str | None = None,
    ) -> OrderCreated:
        json_body = _filter_params(
            market=market,
            side=side,
            volume=volume,
            price=price,
            ord_type=ord_type,
            identifier=identifier,
            time_in_force=time_in_force,
            smp_type=smp_type,
        )
        response = self._transport.request(
            "POST",
            "/v1/orders/test",
            json_body=json_body,
            credentials=self._credentials,
        )
        return OrderCreated.model_validate(response.data)

    def get(
        self,
        *,
        uuid: str | None = None,
        identifier: str | None = None,
    ) -> OrderDetail:
        params = _filter_params(uuid=uuid, identifier=identifier)
        response = self._transport.request(
            "GET", "/v1/order", params=params, credentials=self._credentials
        )
        return OrderDetail.model_validate(response.data)

    def cancel(
        self,
        *,
        uuid: str | None = None,
        identifier: str | None = None,
    ) -> OrderCanceled:
        params = _filter_params(uuid=uuid, identifier=identifier)
        response = self._transport.request(
            "DELETE", "/v1/order", params=params, credentials=self._credentials
        )
        return OrderCanceled.model_validate(response.data)

    def list_open(
        self,
        *,
        market: str | None = None,
        state: str | None = None,
        states: list[str] | None = None,
        page: int | None = None,
        limit: int | None = None,
        order_by: str | None = None,
    ) -> list[OrderOpen]:
        params: dict[str, Any] = _filter_params(
            market=market, state=state, page=page, limit=limit, order_by=order_by
        )
        if states is not None:
            params["states[]"] = states
        response = self._transport.request(
            "GET", "/v1/orders/open", params=params, credentials=self._credentials
        )
        return [OrderOpen.model_validate(item) for item in response.data]

    def cancel_open(
        self,
        *,
        quote_currencies: str | None = None,
        cancel_side: str | None = None,
        count: int | None = None,
        order_by: str | None = None,
        pairs: str | None = None,
        exclude_pairs: str | None = None,
    ) -> CancelResult:
        params = _filter_params(
            quote_currencies=quote_currencies,
            cancel_side=cancel_side,
            count=count,
            order_by=order_by,
            pairs=pairs,
            exclude_pairs=exclude_pairs,
        )
        response = self._transport.request(
            "DELETE", "/v1/orders/open", params=params, credentials=self._credentials
        )
        return CancelResult.model_validate(response.data)

    def list_closed(
        self,
        *,
        market: str | None = None,
        state: str | None = None,
        states: list[str] | None = None,
        start_time: str | None = None,
        end_time: str | None = None,
        limit: int | None = None,
        order_by: str | None = None,
    ) -> list[OrderClosed]:
        params: dict[str, Any] = _filter_params(
            market=market,
            state=state,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            order_by=order_by,
        )
        if states is not None:
            params["states[]"] = states
        response = self._transport.request(
            "GET", "/v1/orders/closed", params=params, credentials=self._credentials
        )
        return [OrderClosed.model_validate(item) for item in response.data]

    def list_by_ids(
        self,
        *,
        market: str | None = None,
        uuids: list[str] | None = None,
        identifiers: list[str] | None = None,
        order_by: str | None = None,
    ) -> list[OrderByIds]:
        params: dict[str, Any] = _filter_params(market=market, order_by=order_by)
        if uuids is not None:
            params["uuids[]"] = uuids
        if identifiers is not None:
            params["identifiers[]"] = identifiers
        response = self._transport.request(
            "GET", "/v1/orders/uuids", params=params, credentials=self._credentials
        )
        return [OrderByIds.model_validate(item) for item in response.data]

    def cancel_by_ids(
        self,
        *,
        uuids: list[str] | None = None,
        identifiers: list[str] | None = None,
    ) -> CancelResult:
        params: dict[str, Any] = {}
        if uuids is not None:
            params["uuids[]"] = uuids
        if identifiers is not None:
            params["identifiers[]"] = identifiers
        response = self._transport.request(
            "DELETE", "/v1/orders/uuids", params=params, credentials=self._credentials
        )
        return CancelResult.model_validate(response.data)

    def cancel_and_new(
        self,
        *,
        new_ord_type: str,
        prev_order_uuid: str | None = None,
        prev_order_identifier: str | None = None,
        new_volume: str | None = None,
        new_price: str | None = None,
        new_identifier: str | None = None,
        new_time_in_force: str | None = None,
        new_smp_type: str | None = None,
    ) -> CancelAndNewOrderResponse:
        json_body = _filter_params(
            new_ord_type=new_ord_type,
            prev_order_uuid=prev_order_uuid,
            prev_order_identifier=prev_order_identifier,
            new_volume=new_volume,
            new_price=new_price,
            new_identifier=new_identifier,
            new_time_in_force=new_time_in_force,
            new_smp_type=new_smp_type,
        )
        response = self._transport.request(
            "POST",
            "/v1/orders/cancel_and_new",
            json_body=json_body,
            credentials=self._credentials,
        )
        return CancelAndNewOrderResponse.model_validate(response.data)

    def get_chance(self, *, market: str) -> list[OrderChance]:
        params = {"market": market}
        response = self._transport.request(
            "GET", "/v1/orders/chance", params=params, credentials=self._credentials
        )
        return [OrderChance.model_validate(item) for item in response.data]


class AsyncOrdersAPI(_AsyncAPIResource):
    async def create(
        self,
        *,
        market: str,
        side: str,
        volume: str | None = None,
        price: str | None = None,
        ord_type: str = "limit",
        identifier: str | None = None,
        time_in_force: str | None = None,
        smp_type: str | None = None,
    ) -> OrderCreated:
        json_body = _filter_params(
            market=market,
            side=side,
            volume=volume,
            price=price,
            ord_type=ord_type,
            identifier=identifier,
            time_in_force=time_in_force,
            smp_type=smp_type,
        )
        response = await self._transport.request(
            "POST", "/v1/orders", json_body=json_body, credentials=self._credentials
        )
        return OrderCreated.model_validate(response.data)

    async def create_test(
        self,
        *,
        market: str,
        side: str,
        volume: str | None = None,
        price: str | None = None,
        ord_type: str = "limit",
        identifier: str | None = None,
        time_in_force: str | None = None,
        smp_type: str | None = None,
    ) -> OrderCreated:
        json_body = _filter_params(
            market=market,
            side=side,
            volume=volume,
            price=price,
            ord_type=ord_type,
            identifier=identifier,
            time_in_force=time_in_force,
            smp_type=smp_type,
        )
        response = await self._transport.request(
            "POST",
            "/v1/orders/test",
            json_body=json_body,
            credentials=self._credentials,
        )
        return OrderCreated.model_validate(response.data)

    async def get(
        self,
        *,
        uuid: str | None = None,
        identifier: str | None = None,
    ) -> OrderDetail:
        params = _filter_params(uuid=uuid, identifier=identifier)
        response = await self._transport.request(
            "GET", "/v1/order", params=params, credentials=self._credentials
        )
        return OrderDetail.model_validate(response.data)

    async def cancel(
        self,
        *,
        uuid: str | None = None,
        identifier: str | None = None,
    ) -> OrderCanceled:
        params = _filter_params(uuid=uuid, identifier=identifier)
        response = await self._transport.request(
            "DELETE", "/v1/order", params=params, credentials=self._credentials
        )
        return OrderCanceled.model_validate(response.data)

    async def list_open(
        self,
        *,
        market: str | None = None,
        state: str | None = None,
        states: list[str] | None = None,
        page: int | None = None,
        limit: int | None = None,
        order_by: str | None = None,
    ) -> list[OrderOpen]:
        params: dict[str, Any] = _filter_params(
            market=market, state=state, page=page, limit=limit, order_by=order_by
        )
        if states is not None:
            params["states[]"] = states
        response = await self._transport.request(
            "GET", "/v1/orders/open", params=params, credentials=self._credentials
        )
        return [OrderOpen.model_validate(item) for item in response.data]

    async def cancel_open(
        self,
        *,
        quote_currencies: str | None = None,
        cancel_side: str | None = None,
        count: int | None = None,
        order_by: str | None = None,
        pairs: str | None = None,
        exclude_pairs: str | None = None,
    ) -> CancelResult:
        params = _filter_params(
            quote_currencies=quote_currencies,
            cancel_side=cancel_side,
            count=count,
            order_by=order_by,
            pairs=pairs,
            exclude_pairs=exclude_pairs,
        )
        response = await self._transport.request(
            "DELETE", "/v1/orders/open", params=params, credentials=self._credentials
        )
        return CancelResult.model_validate(response.data)

    async def list_closed(
        self,
        *,
        market: str | None = None,
        state: str | None = None,
        states: list[str] | None = None,
        start_time: str | None = None,
        end_time: str | None = None,
        limit: int | None = None,
        order_by: str | None = None,
    ) -> list[OrderClosed]:
        params: dict[str, Any] = _filter_params(
            market=market,
            state=state,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            order_by=order_by,
        )
        if states is not None:
            params["states[]"] = states
        response = await self._transport.request(
            "GET", "/v1/orders/closed", params=params, credentials=self._credentials
        )
        return [OrderClosed.model_validate(item) for item in response.data]

    async def list_by_ids(
        self,
        *,
        market: str | None = None,
        uuids: list[str] | None = None,
        identifiers: list[str] | None = None,
        order_by: str | None = None,
    ) -> list[OrderByIds]:
        params: dict[str, Any] = _filter_params(market=market, order_by=order_by)
        if uuids is not None:
            params["uuids[]"] = uuids
        if identifiers is not None:
            params["identifiers[]"] = identifiers
        response = await self._transport.request(
            "GET", "/v1/orders/uuids", params=params, credentials=self._credentials
        )
        return [OrderByIds.model_validate(item) for item in response.data]

    async def cancel_by_ids(
        self,
        *,
        uuids: list[str] | None = None,
        identifiers: list[str] | None = None,
    ) -> CancelResult:
        params: dict[str, Any] = {}
        if uuids is not None:
            params["uuids[]"] = uuids
        if identifiers is not None:
            params["identifiers[]"] = identifiers
        response = await self._transport.request(
            "DELETE", "/v1/orders/uuids", params=params, credentials=self._credentials
        )
        return CancelResult.model_validate(response.data)

    async def cancel_and_new(
        self,
        *,
        new_ord_type: str,
        prev_order_uuid: str | None = None,
        prev_order_identifier: str | None = None,
        new_volume: str | None = None,
        new_price: str | None = None,
        new_identifier: str | None = None,
        new_time_in_force: str | None = None,
        new_smp_type: str | None = None,
    ) -> CancelAndNewOrderResponse:
        json_body = _filter_params(
            new_ord_type=new_ord_type,
            prev_order_uuid=prev_order_uuid,
            prev_order_identifier=prev_order_identifier,
            new_volume=new_volume,
            new_price=new_price,
            new_identifier=new_identifier,
            new_time_in_force=new_time_in_force,
            new_smp_type=new_smp_type,
        )
        response = await self._transport.request(
            "POST",
            "/v1/orders/cancel_and_new",
            json_body=json_body,
            credentials=self._credentials,
        )
        return CancelAndNewOrderResponse.model_validate(response.data)

    async def get_chance(self, *, market: str) -> list[OrderChance]:
        params = {"market": market}
        response = await self._transport.request(
            "GET", "/v1/orders/chance", params=params, credentials=self._credentials
        )
        return [OrderChance.model_validate(item) for item in response.data]



from __future__ import annotations

from collections.abc import Callable
from typing import Literal

from upbeat.strategies._base import (
    AsyncClientProtocol,
    DCAResult,
    PortfolioItem,
    RebalanceOrder,
    RebalanceResult,
    SyncClientProtocol,
)
from upbeat.strategies.alerts import async_price_alert, price_alert
from upbeat.strategies.dca import async_dca, dca
from upbeat.strategies.order_helpers import (
    async_cancel_all_orders,
    async_market_buy_krw,
    async_place_and_wait,
    cancel_all_orders,
    market_buy_krw,
    place_and_wait,
)
from upbeat.strategies.portfolio import (
    async_get_portfolio,
    async_get_portfolio_value,
    get_portfolio,
    get_portfolio_value,
)
from upbeat.strategies.rebalance import async_rebalance, rebalance
from upbeat.types.order import CancelResult, OrderCreated, OrderDetail


class SyncStrategies:
    def __init__(self, client: SyncClientProtocol) -> None:
        self._client = client

    def get_portfolio_value(self) -> float:
        return get_portfolio_value(self._client)

    def get_portfolio(self) -> list[PortfolioItem]:
        return get_portfolio(self._client)

    def market_buy_krw(self, market: str, krw_amount: float) -> OrderCreated:
        return market_buy_krw(self._client, market, krw_amount)

    def place_and_wait(
        self,
        *,
        market: str,
        side: str,
        ord_type: str,
        price: str | None = None,
        volume: str | None = None,
        timeout: float,
        poll_interval: float = 1.0,
    ) -> OrderDetail:
        return place_and_wait(
            self._client,
            market=market,
            side=side,
            ord_type=ord_type,
            price=price,
            volume=volume,
            timeout=timeout,
            poll_interval=poll_interval,
        )

    def cancel_all_orders(self, market: str | None = None) -> CancelResult:
        return cancel_all_orders(self._client, market)

    def dca(
        self,
        *,
        market: str,
        total_amount: float,
        splits: int,
        interval_seconds: float,
    ) -> DCAResult:
        return dca(
            self._client,
            market=market,
            total_amount=total_amount,
            splits=splits,
            interval_seconds=interval_seconds,
        )

    def rebalance(self, *, targets: dict[str, float]) -> RebalanceResult:
        return rebalance(self._client, targets=targets)

    def price_alert(
        self,
        *,
        market: str,
        stop_loss: float | None = None,
        take_profit: float | None = None,
        callback: Callable[[str, Literal["stop_loss", "take_profit"], float], None],
        poll_interval: float = 5.0,
    ) -> None:
        price_alert(
            self._client,
            market=market,
            stop_loss=stop_loss,
            take_profit=take_profit,
            callback=callback,
            poll_interval=poll_interval,
        )


class AsyncStrategies:
    def __init__(self, client: AsyncClientProtocol) -> None:
        self._client = client

    async def get_portfolio_value(self) -> float:
        return await async_get_portfolio_value(self._client)

    async def get_portfolio(self) -> list[PortfolioItem]:
        return await async_get_portfolio(self._client)

    async def market_buy_krw(self, market: str, krw_amount: float) -> OrderCreated:
        return await async_market_buy_krw(self._client, market, krw_amount)

    async def place_and_wait(
        self,
        *,
        market: str,
        side: str,
        ord_type: str,
        price: str | None = None,
        volume: str | None = None,
        timeout: float,
        poll_interval: float = 1.0,
    ) -> OrderDetail:
        return await async_place_and_wait(
            self._client,
            market=market,
            side=side,
            ord_type=ord_type,
            price=price,
            volume=volume,
            timeout=timeout,
            poll_interval=poll_interval,
        )

    async def cancel_all_orders(self, market: str | None = None) -> CancelResult:
        return await async_cancel_all_orders(self._client, market)

    async def dca(
        self,
        *,
        market: str,
        total_amount: float,
        splits: int,
        interval_seconds: float,
    ) -> DCAResult:
        return await async_dca(
            self._client,
            market=market,
            total_amount=total_amount,
            splits=splits,
            interval_seconds=interval_seconds,
        )

    async def rebalance(self, *, targets: dict[str, float]) -> RebalanceResult:
        return await async_rebalance(self._client, targets=targets)

    async def price_alert(
        self,
        *,
        market: str,
        stop_loss: float | None = None,
        take_profit: float | None = None,
        callback: Callable[[str, Literal["stop_loss", "take_profit"], float], None],
        poll_interval: float = 5.0,
    ) -> None:
        await async_price_alert(
            self._client,
            market=market,
            stop_loss=stop_loss,
            take_profit=take_profit,
            callback=callback,
            poll_interval=poll_interval,
        )


__all__ = [
    "AsyncStrategies",
    "DCAResult",
    "PortfolioItem",
    "RebalanceOrder",
    "RebalanceResult",
    "SyncStrategies",
]



from __future__ import annotations

from decimal import Decimal

from upbeat._constants import round_to_tick
from upbeat.strategies._base import (
    AsyncClientProtocol,
    RebalanceOrder,
    RebalanceResult,
    SyncClientProtocol,
)
from upbeat.strategies.portfolio import (
    async_get_portfolio,
    async_get_portfolio_value,
    get_portfolio,
    get_portfolio_value,
)

_MIN_ORDER_KRW = 5_000.0


def rebalance(
    client: SyncClientProtocol,
    *,
    targets: dict[str, float],
) -> RebalanceResult:
    weight_sum = sum(targets.values())
    if abs(weight_sum - 1.0) > 1e-6:
        raise ValueError(f"Target weights must sum to 1.0, got {weight_sum}")

    total_value = get_portfolio_value(client)
    portfolio = get_portfolio(client)

    current_map: dict[str, float] = {}
    price_map: dict[str, float] = {}
    for item in portfolio:
        current_map[item.market] = item.estimated_value_krw
        if item.market != "KRW":
            price_map[item.market] = item.current_price

    all_markets = set(targets.keys()) | set(current_map.keys())

    diffs: list[tuple[str, float]] = []
    for market in all_markets:
        target_value = total_value * targets.get(market, 0.0)
        current_value = current_map.get(market, 0.0)
        diff = target_value - current_value
        diffs.append((market, diff))

    sells = [(m, d) for m, d in diffs if d < -_MIN_ORDER_KRW and m != "KRW"]
    buys = [(m, d) for m, d in diffs if d > _MIN_ORDER_KRW and m != "KRW"]

    result_orders: list[RebalanceOrder] = []

    for market, diff in sells:
        price = price_map[market]
        volume = abs(diff) / price
        order = client.orders.create(
            market=market,
            side="ask",
            ord_type="market",
            volume=str(volume),
        )
        result_orders.append(
            RebalanceOrder(
                market=market,
                side="ask",
                ord_type="market",
                amount=abs(diff),
                order=order,
            )
        )

    for market, diff in buys:
        rounded = round_to_tick(Decimal(str(diff)))
        order = client.orders.create(
            market=market,
            side="bid",
            ord_type="price",
            price=str(int(rounded)),
        )
        result_orders.append(
            RebalanceOrder(
                market=market,
                side="bid",
                ord_type="price",
                amount=diff,
                order=order,
            )
        )

    return RebalanceResult(orders=result_orders, total_value=total_value)


async def async_rebalance(
    client: AsyncClientProtocol,
    *,
    targets: dict[str, float],
) -> RebalanceResult:
    weight_sum = sum(targets.values())
    if abs(weight_sum - 1.0) > 1e-6:
        raise ValueError(f"Target weights must sum to 1.0, got {weight_sum}")

    total_value = await async_get_portfolio_value(client)
    portfolio = await async_get_portfolio(client)

    current_map: dict[str, float] = {}
    price_map: dict[str, float] = {}
    for item in portfolio:
        current_map[item.market] = item.estimated_value_krw
        if item.market != "KRW":
            price_map[item.market] = item.current_price

    all_markets = set(targets.keys()) | set(current_map.keys())

    diffs: list[tuple[str, float]] = []
    for market in all_markets:
        target_value = total_value * targets.get(market, 0.0)
        current_value = current_map.get(market, 0.0)
        diff = target_value - current_value
        diffs.append((market, diff))

    sells = [(m, d) for m, d in diffs if d < -_MIN_ORDER_KRW and m != "KRW"]
    buys = [(m, d) for m, d in diffs if d > _MIN_ORDER_KRW and m != "KRW"]

    result_orders: list[RebalanceOrder] = []

    for market, diff in sells:
        price = price_map[market]
        volume = abs(diff) / price
        order = await client.orders.create(
            market=market,
            side="ask",
            ord_type="market",
            volume=str(volume),
        )
        result_orders.append(
            RebalanceOrder(
                market=market,
                side="ask",
                ord_type="market",
                amount=abs(diff),
                order=order,
            )
        )

    for market, diff in buys:
        rounded = round_to_tick(Decimal(str(diff)))
        order = await client.orders.create(
            market=market,
            side="bid",
            ord_type="price",
            price=str(int(rounded)),
        )
        result_orders.append(
            RebalanceOrder(
                market=market,
                side="bid",
                ord_type="price",
                amount=diff,
                order=order,
            )
        )

    return RebalanceResult(orders=result_orders, total_value=total_value)



from __future__ import annotations

import asyncio
import json
from collections.abc import AsyncIterator
from typing import TypeVar

import websockets
import websockets.asyncio.client
import websockets.exceptions
from pydantic import BaseModel

from upbeat._auth import Credentials, build_auth_header
from upbeat._constants import WS_PRIVATE_URL, WS_PUBLIC_URL
from upbeat._errors import (
    WebSocketClosedError,
    WebSocketConnectionError,
    WebSocketError,
)
from upbeat._logger import Logger, WebSocketEventInfo
from upbeat.ws._payload import generate_payload
from upbeat.ws.types import (
    CandleMessage,
    MyAssetMessage,
    MyOrderMessage,
    OrderbookMessage,
    TickerMessage,
    TradeMessage,
    WebSocketMessage,
    parse_message,
)

_T = TypeVar("_T", bound=BaseModel)

_MAX_RECONNECT_DELAY = 60.0
_INITIAL_RECONNECT_DELAY = 1.0


class AsyncUpbeatWebSocket:
    def __init__(
        self,
        credentials: Credentials | None,
        logger: Logger | None,
    ) -> None:
        self._credentials = credentials
        self._logger = logger

    # ── Single-channel convenience methods ────────────────────────────

    async def ticker(
        self,
        codes: list[str],
        *,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> AsyncIterator[TickerMessage]:
        payload = generate_payload(
            type="ticker",
            codes=codes,
            is_only_snapshot=is_only_snapshot,
            is_only_realtime=is_only_realtime,
        )
        async for msg in self._stream(WS_PUBLIC_URL, payload, TickerMessage):
            yield msg

    async def trade(
        self,
        codes: list[str],
        *,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> AsyncIterator[TradeMessage]:
        payload = generate_payload(
            type="trade",
            codes=codes,
            is_only_snapshot=is_only_snapshot,
            is_only_realtime=is_only_realtime,
        )
        async for msg in self._stream(WS_PUBLIC_URL, payload, TradeMessage):
            yield msg

    async def orderbook(
        self,
        codes: list[str],
        *,
        level: float | None = None,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> AsyncIterator[OrderbookMessage]:
        payload = generate_payload(
            type="orderbook",
            codes=codes,
            level=level,
            is_only_snapshot=is_only_snapshot,
            is_only_realtime=is_only_realtime,
        )
        async for msg in self._stream(WS_PUBLIC_URL, payload, OrderbookMessage):
            yield msg

    async def candle(
        self,
        codes: list[str],
        *,
        type: str = "candle.1m",
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> AsyncIterator[CandleMessage]:
        payload = generate_payload(
            type=type,
            codes=codes,
            is_only_snapshot=is_only_snapshot,
            is_only_realtime=is_only_realtime,
        )
        async for msg in self._stream(WS_PUBLIC_URL, payload, CandleMessage):
            yield msg

    async def my_orders(
        self,
        codes: list[str] | None = None,
    ) -> AsyncIterator[MyOrderMessage]:
        payload = generate_payload(type="myOrder", codes=codes)
        async for msg in self._stream(
            WS_PRIVATE_URL, payload, MyOrderMessage, is_private=True
        ):
            yield msg

    async def my_assets(self) -> AsyncIterator[MyAssetMessage]:
        payload = generate_payload(type="myAsset")
        async for msg in self._stream(
            WS_PRIVATE_URL, payload, MyAssetMessage, is_private=True
        ):
            yield msg

    # ── connect() — multi-channel context manager ─────────────────────

    def connect(self, *, private: bool = False) -> WebSocketConnection:
        return WebSocketConnection(
            credentials=self._credentials,
            logger=self._logger,
            private=private,
        )

    # ── Core stream loop with auto-reconnect ──────────────────────────

    async def _stream(
        self,
        url: str,
        payload: list[dict[str, object]],
        model: type[_T],
        *,
        is_private: bool = False,
    ) -> AsyncIterator[_T]:
        if is_private and self._credentials is None:
            raise WebSocketError("Credentials required for private WebSocket channels")

        delay = _INITIAL_RECONNECT_DELAY
        while True:
            headers: dict[str, str] = {}
            if is_private and self._credentials is not None:
                headers["Authorization"] = build_auth_header(self._credentials)

            try:
                async with websockets.asyncio.client.connect(
                    url,
                    additional_headers=headers,
                    ping_interval=60,
                    ping_timeout=20,
                ) as ws:
                    await ws.send(json.dumps(payload))
                    self._log("connect", url)
                    delay = _INITIAL_RECONNECT_DELAY

                    async for raw in ws:
                        msg = self._process_raw(raw, model)
                        if msg is not None:
                            yield msg

            except websockets.exceptions.ConnectionClosed as exc:
                self._log("disconnect", url, error=exc)
                await asyncio.sleep(delay)
                delay = min(delay * 2, _MAX_RECONNECT_DELAY)
                continue

            except OSError as exc:
                self._log("connection_error", url, error=exc)
                raise WebSocketConnectionError(str(exc)) from exc

    # ── Helpers ───────────────────────────────────────────────────────

    def _process_raw(self, raw: str | bytes, model: type[_T]) -> _T | None:
        text = raw.decode("utf-8") if isinstance(raw, bytes) else raw

        if text.strip() == '"UP"' or text.strip() == "UP":
            return None

        data = json.loads(text)

        if isinstance(data, dict) and "error" in data:
            error_obj = data["error"]
            name = error_obj.get("name", "UNKNOWN")
            message = error_obj.get("message", "")
            raise WebSocketError(f"{name}: {message}")

        return model.model_validate(data)

    def _log(
        self,
        event: str,
        url: str,
        *,
        error: Exception | None = None,
    ) -> None:
        if self._logger is not None:
            self._logger.on_ws_event(
                WebSocketEventInfo(event=event, url=url, error=error)
            )


# ── WebSocketConnection — multi-channel context manager ───────────────


class WebSocketConnection:
    def __init__(
        self,
        *,
        credentials: Credentials | None,
        logger: Logger | None,
        private: bool,
    ) -> None:
        self._credentials = credentials
        self._logger = logger
        self._private = private
        self._subscriptions: list[dict[str, object]] = []
        self._ws: websockets.asyncio.client.ClientConnection | None = None
        self._payload_sent = False

    # ── Subscribe methods (synchronous registration) ──────────────────

    def subscribe_ticker(
        self,
        codes: list[str],
        *,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> None:
        sub: dict[str, object] = {"type": "ticker", "codes": codes}
        if is_only_snapshot:
            sub["is_only_snapshot"] = True
        if is_only_realtime:
            sub["is_only_realtime"] = True
        self._subscriptions.append(sub)

    def subscribe_trade(
        self,
        codes: list[str],
        *,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> None:
        sub: dict[str, object] = {"type": "trade", "codes": codes}
        if is_only_snapshot:
            sub["is_only_snapshot"] = True
        if is_only_realtime:
            sub["is_only_realtime"] = True
        self._subscriptions.append(sub)

    def subscribe_orderbook(
        self,
        codes: list[str],
        *,
        level: float | None = None,
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> None:
        sub: dict[str, object] = {"type": "orderbook", "codes": codes}
        if level is not None:
            sub["level"] = level
        if is_only_snapshot:
            sub["is_only_snapshot"] = True
        if is_only_realtime:
            sub["is_only_realtime"] = True
        self._subscriptions.append(sub)

    def subscribe_candle(
        self,
        codes: list[str],
        *,
        type: str = "candle.1m",
        is_only_snapshot: bool = False,
        is_only_realtime: bool = False,
    ) -> None:
        sub: dict[str, object] = {"type": type, "codes": codes}
        if is_only_snapshot:
            sub["is_only_snapshot"] = True
        if is_only_realtime:
            sub["is_only_realtime"] = True
        self._subscriptions.append(sub)

    def subscribe_my_orders(self, codes: list[str] | None = None) -> None:
        sub: dict[str, object] = {"type": "myOrder"}
        if codes is not None:
            sub["codes"] = codes
        self._subscriptions.append(sub)

    def subscribe_my_assets(self) -> None:
        self._subscriptions.append({"type": "myAsset"})

    # ── Async context manager ─────────────────────────────────────────

    async def __aenter__(self) -> WebSocketConnection:
        if self._private and self._credentials is None:
            raise WebSocketError("Credentials required for private WebSocket channels")

        url = WS_PRIVATE_URL if self._private else WS_PUBLIC_URL
        headers: dict[str, str] = {}
        if self._private and self._credentials is not None:
            headers["Authorization"] = build_auth_header(self._credentials)

        self._ws = await websockets.asyncio.client.connect(
            url,
            additional_headers=headers,
            ping_interval=60,
            ping_timeout=20,
        ).__aenter__()

        self._log("connect", url)
        return self

    async def __aexit__(self, *_args: object) -> None:
        if self._ws is not None:
            url = WS_PRIVATE_URL if self._private else WS_PUBLIC_URL
            await self._ws.close()
            self._log("disconnect", url)
            self._ws = None

    # ── Async iterator ────────────────────────────────────────────────

    def __aiter__(self) -> WebSocketConnection:
        return self

    async def __anext__(self) -> WebSocketMessage:
        if self._ws is None:
            raise StopAsyncIteration

        if not self._payload_sent:
            await self._send_subscriptions()
            self._payload_sent = True

        while True:
            try:
                raw = await self._ws.recv()
            except websockets.exceptions.ConnectionClosed as exc:
                code = exc.rcvd.code if exc.rcvd else None
                reason = str(exc.rcvd.reason) if exc.rcvd else None
                raise WebSocketClosedError(str(exc), code=code, reason=reason) from exc

            text = raw.decode("utf-8") if isinstance(raw, bytes) else raw

            if text.strip() == '"UP"' or text.strip() == "UP":
                continue

            data = json.loads(text)

            if isinstance(data, dict) and "error" in data:
                error_obj = data["error"]
                name = error_obj.get("name", "UNKNOWN")
                message = error_obj.get("message", "")
                raise WebSocketError(f"{name}: {message}")

            return parse_message(data)

    # ── Internal ──────────────────────────────────────────────────────

    async def _send_subscriptions(self) -> None:
        if self._ws is None or not self._subscriptions:
            return

        import uuid

        payload: list[dict[str, object]] = [{"ticket": str(uuid.uuid4())}]
        payload.extend(self._subscriptions)
        payload.append({"format": "DEFAULT"})
        await self._ws.send(json.dumps(payload))

    def _log(
        self,
        event: str,
        url: str,
        *,
        error: Exception | None = None,
    ) -> None:
        if self._logger is not None:
            self._logger.on_ws_event(
                WebSocketEventInfo(event=event, url=url, error=error)
            )



from upbeat.types.account import Account
from upbeat.types.common import APIResponse
from upbeat.types.deposit import (
    Deposit,
    DepositAddress,
    DepositAddressCreated,
    DepositChanceCoin,
)
from upbeat.types.market import MarketCaution, MarketEvent, TradingPair
from upbeat.types.order import (
    CancelAndNewOrderResponse,
    CancelResult,
    CancelResultGroup,
    CancelResultOrder,
    OrderByIds,
    OrderCanceled,
    OrderChance,
    OrderChanceAccount,
    OrderChanceMarket,
    OrderChanceMarketBidAsk,
    OrderClosed,
    OrderCreated,
    OrderDetail,
    OrderOpen,
    OrderTrade,
)
from upbeat.types.quotation import (
    CandleDay,
    CandleMinute,
    CandlePeriod,
    CandleSecond,
    Orderbook,
    OrderbookInstrument,
    OrderbookUnit,
    Ticker,
    Trade,
)
from upbeat.types.travel_rule import TravelRuleVasp, TravelRuleVerification
from upbeat.types.withdrawal import (
    Withdrawal,
    WithdrawalAddress,
    WithdrawalChance,
    WithdrawalChanceAccount,
    WithdrawalChanceCurrency,
    WithdrawalChanceMemberLevel,
    WithdrawalChanceWithdrawLimit,
    WithdrawalKrw,
)

__all__ = [
    "APIResponse",
    "Account",
    "CancelAndNewOrderResponse",
    "Deposit",
    "DepositAddress",
    "DepositAddressCreated",
    "DepositChanceCoin",
    "CancelResult",
    "CancelResultGroup",
    "CancelResultOrder",
    "CandleDay",
    "CandleMinute",
    "CandlePeriod",
    "CandleSecond",
    "MarketCaution",
    "MarketEvent",
    "Orderbook",
    "OrderbookInstrument",
    "OrderbookUnit",
    "OrderByIds",
    "OrderCanceled",
    "OrderChance",
    "OrderChanceAccount",
    "OrderChanceMarket",
    "OrderChanceMarketBidAsk",
    "OrderClosed",
    "OrderCreated",
    "OrderDetail",
    "OrderOpen",
    "OrderTrade",
    "Ticker",
    "Trade",
    "TradingPair",
    "TravelRuleVasp",
    "TravelRuleVerification",
    "Withdrawal",
    "WithdrawalAddress",
    "WithdrawalChance",
    "WithdrawalChanceAccount",
    "WithdrawalChanceCurrency",
    "WithdrawalChanceMemberLevel",
    "WithdrawalChanceWithdrawLimit",
    "WithdrawalKrw",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from upbeat.shortcuts._accounts import find_account

if TYPE_CHECKING:
    from upbeat._client import AsyncUpbeat, Upbeat
    from upbeat.types.account import Account


class SyncShortcuts:
    def __init__(self, client: Upbeat) -> None:
        self._client = client

    def get_account(self, currency: str) -> Account | None:
        return find_account(self._client.accounts.list(), currency)


class AsyncShortcuts:
    def __init__(self, client: AsyncUpbeat) -> None:
        self._client = client

    async def get_account(self, currency: str) -> Account | None:
        accounts = await self._client.accounts.list()
        return find_account(accounts, currency)


__all__ = [
    "AsyncShortcuts",
    "SyncShortcuts",
    "find_account",
]



from importlib.metadata import version as _version

from upbeat._auth import Credentials
from upbeat._client import AsyncUpbeat, Upbeat
from upbeat._config import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT, Timeout, UpbeatConfig
from upbeat._constants import (
    API_BASE_URL,
    WS_PRIVATE_URL,
    WS_PUBLIC_URL,
    MarketState,
    OrderSide,
    OrderState,
    OrderType,
    SortOrder,
    TimeInForce,
    get_krw_tick_size,
    round_to_tick,
)
from upbeat._convenience import (
    get_candles,
    get_candles_df,
    get_markets,
    get_orderbook,
    get_orderbooks,
    get_ticker,
    get_tickers,
    get_trades,
)
from upbeat._errors import (
    APIConnectionError,
    APIError,
    APIStatusError,
    APITimeoutError,
    AuthenticationError,
    BadRequestError,
    InsufficientFundsError,
    InternalServerError,
    MinimumOrderError,
    NotFoundError,
    PermissionDeniedError,
    RateLimitError,
    RemainingRequest,
    UnprocessableEntityError,
    UpbeatError,
    WebSocketClosedError,
    WebSocketConnectionError,
    WebSocketError,
)
from upbeat._logger import (
    DefaultLogger,
    ErrorInfo,
    Logger,
    NullLogger,
    RequestInfo,
    ResponseInfo,
    RetryInfo,
    WebSocketEventInfo,
)
from upbeat.types.common import APIResponse
from upbeat.ws import (
    AsyncUpbeatWebSocket,
    CandleMessage,
    MyAssetItem,
    MyAssetMessage,
    MyOrderMessage,
    OrderbookMessage,
    OrderbookUnitMessage,
    TickerMessage,
    TradeMessage,
    WebSocketConnection,
    WebSocketMessage,
    generate_multi_payload,
    generate_payload,
    parse_message,
)

__version__ = _version("upbeat")

__all__ = [
    # Version
    "__version__",
    # Client
    "AsyncUpbeat",
    "Upbeat",
    # Auth
    "Credentials",
    # Types
    "APIResponse",
    # Config
    "DEFAULT_MAX_RETRIES",
    "DEFAULT_TIMEOUT",
    "Timeout",
    "UpbeatConfig",
    # Constants
    "API_BASE_URL",
    "WS_PRIVATE_URL",
    "WS_PUBLIC_URL",
    "MarketState",
    "OrderSide",
    "OrderState",
    "OrderType",
    "SortOrder",
    "TimeInForce",
    "get_krw_tick_size",
    "round_to_tick",
    # Logger
    "DefaultLogger",
    "ErrorInfo",
    "Logger",
    "NullLogger",
    "RequestInfo",
    "ResponseInfo",
    "RetryInfo",
    "WebSocketEventInfo",
    # Convenience functions
    "get_candles",
    "get_candles_df",
    "get_markets",
    "get_orderbook",
    "get_orderbooks",
    "get_ticker",
    "get_tickers",
    "get_trades",
    # Errors
    "APIConnectionError",
    "APIError",
    "APIStatusError",
    "APITimeoutError",
    "AuthenticationError",
    "BadRequestError",
    "InsufficientFundsError",
    "InternalServerError",
    "MinimumOrderError",
    "NotFoundError",
    "PermissionDeniedError",
    "RateLimitError",
    "RemainingRequest",
    "UnprocessableEntityError",
    "UpbeatError",
    "WebSocketClosedError",
    "WebSocketConnectionError",
    "WebSocketError",
    # WebSocket
    "AsyncUpbeatWebSocket",
    "CandleMessage",
    "MyAssetItem",
    "MyAssetMessage",
    "MyOrderMessage",
    "OrderbookMessage",
    "OrderbookUnitMessage",
    "TickerMessage",
    "TradeMessage",
    "WebSocketConnection",
    "WebSocketMessage",
    "WS_PRIVATE_URL",
    "WS_PUBLIC_URL",
    "generate_multi_payload",
    "generate_payload",
    "parse_message",
]



from __future__ import annotations

from typing import Any

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.deposit import (
    Deposit,
    DepositAddress,
    DepositAddressCreated,
    DepositChanceCoin,
)


def _filter_params(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}


class DepositsAPI(_SyncAPIResource):
    def get(
        self,
        *,
        uuid: str | None = None,
        txid: str | None = None,
        currency: str | None = None,
    ) -> Deposit:
        params = _filter_params(uuid=uuid, txid=txid, currency=currency)
        response = self._transport.request(
            "GET", "/v1/deposit", params=params, credentials=self._credentials
        )
        return Deposit.model_validate(response.data)

    def list(
        self,
        *,
        currency: str | None = None,
        state: str | None = None,
        uuids: list[str] | None = None,
        txids: list[str] | None = None,
        limit: int | None = None,
        page: int | None = None,
        order_by: str | None = None,
        from_cursor: str | None = None,
        to: str | None = None,
    ) -> list[Deposit]:
        params: dict[str, Any] = _filter_params(
            currency=currency,
            state=state,
            limit=limit,
            page=page,
            order_by=order_by,
            to=to,
        )
        if from_cursor is not None:
            params["from"] = from_cursor
        if uuids is not None:
            params["uuids[]"] = uuids
        if txids is not None:
            params["txids[]"] = txids
        response = self._transport.request(
            "GET", "/v1/deposits", params=params, credentials=self._credentials
        )
        return [Deposit.model_validate(item) for item in response.data]

    def generate_coin_address(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositAddress | DepositAddressCreated:
        json_body = {"currency": currency, "net_type": net_type}
        response = self._transport.request(
            "POST",
            "/v1/deposits/generate_coin_address",
            json_body=json_body,
            credentials=self._credentials,
        )
        if response.status_code == 201:
            return DepositAddressCreated.model_validate(response.data)
        return DepositAddress.model_validate(response.data)

    def get_coin_address(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositAddress:
        params = {"currency": currency, "net_type": net_type}
        response = self._transport.request(
            "GET",
            "/v1/deposits/coin_address",
            params=params,
            credentials=self._credentials,
        )
        return DepositAddress.model_validate(response.data)

    def list_coin_addresses(self) -> list[DepositAddress]:
        response = self._transport.request(
            "GET",
            "/v1/deposits/coin_addresses",
            credentials=self._credentials,
        )
        return [DepositAddress.model_validate(item) for item in response.data]

    def create_krw(
        self,
        *,
        amount: str,
        two_factor_type: str,
    ) -> Deposit:
        json_body = {"amount": amount, "two_factor_type": two_factor_type}
        response = self._transport.request(
            "POST",
            "/v1/deposits/krw",
            json_body=json_body,
            credentials=self._credentials,
        )
        return Deposit.model_validate(response.data)

    def get_chance_coin(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositChanceCoin:
        params = {"currency": currency, "net_type": net_type}
        response = self._transport.request(
            "GET",
            "/v1/deposits/chance/coin",
            params=params,
            credentials=self._credentials,
        )
        return DepositChanceCoin.model_validate(response.data)


class AsyncDepositsAPI(_AsyncAPIResource):
    async def get(
        self,
        *,
        uuid: str | None = None,
        txid: str | None = None,
        currency: str | None = None,
    ) -> Deposit:
        params = _filter_params(uuid=uuid, txid=txid, currency=currency)
        response = await self._transport.request(
            "GET", "/v1/deposit", params=params, credentials=self._credentials
        )
        return Deposit.model_validate(response.data)

    async def list(
        self,
        *,
        currency: str | None = None,
        state: str | None = None,
        uuids: list[str] | None = None,
        txids: list[str] | None = None,
        limit: int | None = None,
        page: int | None = None,
        order_by: str | None = None,
        from_cursor: str | None = None,
        to: str | None = None,
    ) -> list[Deposit]:
        params: dict[str, Any] = _filter_params(
            currency=currency,
            state=state,
            limit=limit,
            page=page,
            order_by=order_by,
            to=to,
        )
        if from_cursor is not None:
            params["from"] = from_cursor
        if uuids is not None:
            params["uuids[]"] = uuids
        if txids is not None:
            params["txids[]"] = txids
        response = await self._transport.request(
            "GET", "/v1/deposits", params=params, credentials=self._credentials
        )
        return [Deposit.model_validate(item) for item in response.data]

    async def generate_coin_address(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositAddress | DepositAddressCreated:
        json_body = {"currency": currency, "net_type": net_type}
        response = await self._transport.request(
            "POST",
            "/v1/deposits/generate_coin_address",
            json_body=json_body,
            credentials=self._credentials,
        )
        if response.status_code == 201:
            return DepositAddressCreated.model_validate(response.data)
        return DepositAddress.model_validate(response.data)

    async def get_coin_address(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositAddress:
        params = {"currency": currency, "net_type": net_type}
        response = await self._transport.request(
            "GET",
            "/v1/deposits/coin_address",
            params=params,
            credentials=self._credentials,
        )
        return DepositAddress.model_validate(response.data)

    async def list_coin_addresses(self) -> list[DepositAddress]:
        response = await self._transport.request(
            "GET",
            "/v1/deposits/coin_addresses",
            credentials=self._credentials,
        )
        return [DepositAddress.model_validate(item) for item in response.data]

    async def create_krw(
        self,
        *,
        amount: str,
        two_factor_type: str,
    ) -> Deposit:
        json_body = {"amount": amount, "two_factor_type": two_factor_type}
        response = await self._transport.request(
            "POST",
            "/v1/deposits/krw",
            json_body=json_body,
            credentials=self._credentials,
        )
        return Deposit.model_validate(response.data)

    async def get_chance_coin(
        self,
        *,
        currency: str,
        net_type: str,
    ) -> DepositChanceCoin:
        params = {"currency": currency, "net_type": net_type}
        response = await self._transport.request(
            "GET",
            "/v1/deposits/chance/coin",
            params=params,
            credentials=self._credentials,
        )
        return DepositChanceCoin.model_validate(response.data)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class Withdrawal(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: str
    uuid: str
    currency: str
    net_type: str | None = None
    txid: str | None = None
    state: str
    created_at: str
    done_at: str | None = None
    amount: str
    fee: str
    transaction_type: str
    is_cancelable: bool


class WithdrawalKrw(BaseModel):
    model_config = ConfigDict(frozen=True)

    type: str
    uuid: str
    currency: str
    txid: str | None = None
    state: str
    created_at: str
    done_at: str | None = None
    amount: str
    fee: str
    transaction_type: str
    is_cancelable: bool


class WithdrawalAddress(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    net_type: str
    network_name: str
    withdraw_address: str
    secondary_address: str | None = None
    beneficiary_name: str | None = None
    beneficiary_company_name: str | None = None
    beneficiary_type: str | None = None
    exchange_name: str | None = None
    wallet_type: str | None = None


class WithdrawalChanceMemberLevel(BaseModel):
    model_config = ConfigDict(frozen=True)

    security_level: int
    fee_level: int
    email_verified: bool
    identity_auth_verified: bool
    bank_account_verified: bool
    two_factor_auth_verified: bool
    locked: bool
    wallet_locked: bool


class WithdrawalChanceCurrency(BaseModel):
    model_config = ConfigDict(frozen=True)

    code: str
    withdraw_fee: str
    is_coin: bool
    wallet_state: str
    wallet_support: list[str]


class WithdrawalChanceAccount(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    balance: str
    locked: str
    avg_buy_price: str
    avg_buy_price_modified: bool
    unit_currency: str


class WithdrawalChanceWithdrawLimit(BaseModel):
    model_config = ConfigDict(frozen=True)

    currency: str
    onetime: str
    daily: str | None
    remaining_daily: str
    remaining_daily_fiat: str
    fiat_currency: str
    minimum: str | None = None
    fixed: int | None = None
    withdraw_delayed_fiat: str | None = None
    can_withdraw: bool | None = None
    remaining_daily_krw: str | None = None


class WithdrawalChance(BaseModel):
    model_config = ConfigDict(frozen=True)

    member_level: WithdrawalChanceMemberLevel
    currency: WithdrawalChanceCurrency
    account: WithdrawalChanceAccount
    withdraw_limit: WithdrawalChanceWithdrawLimit



import asyncio
import random
import time
from typing import Any

import httpx

from upbeat._auth import Credentials, build_auth_header, build_query_string
from upbeat._config import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT, Timeout
from upbeat._constants import API_BASE_URL
from upbeat._errors import (
    APIConnectionError,
    APIStatusError,
    APITimeoutError,
    InternalServerError,
    RateLimitError,
    RemainingRequest,
)
from upbeat._logger import (
    DefaultLogger,
    ErrorInfo,
    Logger,
    RequestInfo,
    ResponseInfo,
    RetryInfo,
)
from upbeat.types.common import APIResponse

# ── Helper functions ─────────────────────────────────────────────────────


def _should_retry(error: Exception) -> bool:
    if isinstance(error, (RateLimitError, InternalServerError)):
        return True
    return isinstance(error, (APIConnectionError, APITimeoutError))


def _calculate_delay(attempt: int, base: float = 0.5, max_delay: float = 8.0) -> float:
    delay = min(base * (2**attempt), max_delay)
    jitter = delay * random.uniform(0, 0.25)  # noqa: S311
    return delay + jitter


def _calculate_rate_limit_delay(error: RateLimitError) -> float:
    if (
        error.remaining_request is not None
        and error.remaining_request.remaining_sec <= 0
    ):
        return 1.0
    return 1.0


# ── RateLimitTracker ─────────────────────────────────────────────────────


class RateLimitTracker:
    def __init__(self, *, auto_throttle: bool = True) -> None:
        self._state: dict[str, RemainingRequest] = {}
        self._auto_throttle = auto_throttle

    def update(self, remaining: RemainingRequest | None) -> None:
        if remaining is not None:
            self._state[remaining.group] = remaining

    def get_delay(self) -> float:
        if not self._auto_throttle:
            return 0.0
        for remaining in self._state.values():
            if remaining.remaining_sec <= 0:
                return 1.0
        return 0.0


# ── SyncTransport ────────────────────────────────────────────────────────


class SyncTransport:
    def __init__(
        self,
        *,
        base_url: str = API_BASE_URL,
        timeout: Timeout = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        auto_throttle: bool = True,
        logger: Logger | None = None,
        http_client: httpx.Client | None = None,
        event_hooks: dict[str, list[Any]] | None = None,
    ) -> None:
        self._base_url = base_url
        self._max_retries = max_retries
        self._logger: Logger = logger or DefaultLogger()
        self._rate_limiter = RateLimitTracker(auto_throttle=auto_throttle)

        if http_client is not None:
            self._client = http_client
            self._owns_client = False
        else:
            client_kwargs: dict[str, Any] = {
                "base_url": base_url,
                "timeout": httpx.Timeout(timeout.read, connect=timeout.connect),
            }
            if event_hooks is not None:
                client_kwargs["event_hooks"] = event_hooks
            self._client = httpx.Client(**client_kwargs)
            self._owns_client = True

    def request(
        self,
        method: str,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        json_body: dict[str, Any] | None = None,
        credentials: Credentials | None = None,
    ) -> APIResponse:
        last_error: Exception | None = None

        for attempt in range(1 + self._max_retries):
            # Proactive throttle
            delay = self._rate_limiter.get_delay()
            if delay > 0:
                time.sleep(delay)

            # Build headers
            headers: dict[str, str] = {}
            if credentials is not None:
                if json_body is not None:
                    query_string = build_query_string(json_body)
                elif params is not None:
                    query_string = build_query_string(params)
                else:
                    query_string = ""
                headers["Authorization"] = build_auth_header(credentials, query_string)

            if json_body is not None:
                headers["Content-Type"] = "application/json; charset=utf-8"

            # Build URL
            url = path if path.startswith("http") else f"{self._base_url}{path}"

            self._logger.on_request(
                RequestInfo(method=method, url=url, headers=headers)
            )

            try:
                start = time.monotonic()
                response = self._client.request(
                    method,
                    url,
                    params=params,
                    json=json_body,
                    headers=headers,
                )
                elapsed_ms = (time.monotonic() - start) * 1000

                self._logger.on_response(
                    ResponseInfo(
                        status_code=response.status_code,
                        url=str(response.url),
                        headers=dict(response.headers),
                        elapsed_ms=elapsed_ms,
                    )
                )

                # Parse Remaining-Req header
                remaining_request = _parse_remaining_req(response)
                self._rate_limiter.update(remaining_request)

                # Error responses
                if response.status_code >= 400:
                    error = APIStatusError.from_response(response)
                    if _should_retry(error) and attempt < self._max_retries:
                        if isinstance(error, RateLimitError):
                            retry_delay = _calculate_rate_limit_delay(error)
                        else:
                            retry_delay = _calculate_delay(attempt)
                        self._logger.on_retry(
                            RetryInfo(
                                attempt=attempt + 1,
                                max_retries=self._max_retries,
                                delay_seconds=retry_delay,
                                reason=str(error),
                                method=method,
                                url=url,
                            )
                        )
                        time.sleep(retry_delay)
                        last_error = error
                        continue
                    self._logger.on_error(
                        ErrorInfo(error=error, method=method, url=url)
                    )
                    raise error

                # Success
                return APIResponse(
                    data=response.json(),
                    remaining_request=remaining_request,
                    status_code=response.status_code,
                )

            except httpx.TimeoutException as exc:
                last_error = APITimeoutError(f"Request timed out: {exc}", cause=exc)
                if attempt < self._max_retries:
                    retry_delay = _calculate_delay(attempt)
                    self._logger.on_retry(
                        RetryInfo(
                            attempt=attempt + 1,
                            max_retries=self._max_retries,
                            delay_seconds=retry_delay,
                            reason=str(last_error),
                            method=method,
                            url=url,
                        )
                    )
                    time.sleep(retry_delay)
                    continue
                self._logger.on_error(
                    ErrorInfo(error=last_error, method=method, url=url)
                )
                raise last_error from exc

            except (
                httpx.ConnectError,
                httpx.NetworkError,
                httpx.RemoteProtocolError,
            ) as exc:
                last_error = APIConnectionError(f"Connection error: {exc}", cause=exc)
                if attempt < self._max_retries:
                    retry_delay = _calculate_delay(attempt)
                    self._logger.on_retry(
                        RetryInfo(
                            attempt=attempt + 1,
                            max_retries=self._max_retries,
                            delay_seconds=retry_delay,
                            reason=str(last_error),
                            method=method,
                            url=url,
                        )
                    )
                    time.sleep(retry_delay)
                    continue
                self._logger.on_error(
                    ErrorInfo(error=last_error, method=method, url=url)
                )
                raise last_error from exc

        # Should not reach here, but just in case
        assert last_error is not None
        raise last_error

    def close(self) -> None:
        if self._owns_client:
            self._client.close()


# ── AsyncTransport ───────────────────────────────────────────────────────


class AsyncTransport:
    def __init__(
        self,
        *,
        base_url: str = API_BASE_URL,
        timeout: Timeout = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        auto_throttle: bool = True,
        logger: Logger | None = None,
        http_client: httpx.AsyncClient | None = None,
        event_hooks: dict[str, list[Any]] | None = None,
    ) -> None:
        self._base_url = base_url
        self._max_retries = max_retries
        self._logger: Logger = logger or DefaultLogger()
        self._rate_limiter = RateLimitTracker(auto_throttle=auto_throttle)

        if http_client is not None:
            self._client = http_client
            self._owns_client = False
        else:
            client_kwargs: dict[str, Any] = {
                "base_url": base_url,
                "timeout": httpx.Timeout(timeout.read, connect=timeout.connect),
            }
            if event_hooks is not None:
                client_kwargs["event_hooks"] = event_hooks
            self._client = httpx.AsyncClient(**client_kwargs)
            self._owns_client = True

    async def request(
        self,
        method: str,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        json_body: dict[str, Any] | None = None,
        credentials: Credentials | None = None,
    ) -> APIResponse:
        last_error: Exception | None = None

        for attempt in range(1 + self._max_retries):
            delay = self._rate_limiter.get_delay()
            if delay > 0:
                await asyncio.sleep(delay)

            headers: dict[str, str] = {}
            if credentials is not None:
                if json_body is not None:
                    query_string = build_query_string(json_body)
                elif params is not None:
                    query_string = build_query_string(params)
                else:
                    query_string = ""
                headers["Authorization"] = build_auth_header(credentials, query_string)

            if json_body is not None:
                headers["Content-Type"] = "application/json; charset=utf-8"

            url = path if path.startswith("http") else f"{self._base_url}{path}"

            self._logger.on_request(
                RequestInfo(method=method, url=url, headers=headers)
            )

            try:
                start = time.monotonic()
                response = await self._client.request(
                    method,
                    url,
                    params=params,
                    json=json_body,
                    headers=headers,
                )
                elapsed_ms = (time.monotonic() - start) * 1000

                self._logger.on_response(
                    ResponseInfo(
                        status_code=response.status_code,
                        url=str(response.url),
                        headers=dict(response.headers),
                        elapsed_ms=elapsed_ms,
                    )
                )

                remaining_request = _parse_remaining_req(response)
                self._rate_limiter.update(remaining_request)

                if response.status_code >= 400:
                    error = APIStatusError.from_response(response)
                    if _should_retry(error) and attempt < self._max_retries:
                        if isinstance(error, RateLimitError):
                            retry_delay = _calculate_rate_limit_delay(error)
                        else:
                            retry_delay = _calculate_delay(attempt)
                        self._logger.on_retry(
                            RetryInfo(
                                attempt=attempt + 1,
                                max_retries=self._max_retries,
                                delay_seconds=retry_delay,
                                reason=str(error),
                                method=method,
                                url=url,
                            )
                        )
                        await asyncio.sleep(retry_delay)
                        last_error = error
                        continue
                    self._logger.on_error(
                        ErrorInfo(error=error, method=method, url=url)
                    )
                    raise error

                return APIResponse(
                    data=response.json(),
                    remaining_request=remaining_request,
                    status_code=response.status_code,
                )

            except httpx.TimeoutException as exc:
                last_error = APITimeoutError(f"Request timed out: {exc}", cause=exc)
                if attempt < self._max_retries:
                    retry_delay = _calculate_delay(attempt)
                    self._logger.on_retry(
                        RetryInfo(
                            attempt=attempt + 1,
                            max_retries=self._max_retries,
                            delay_seconds=retry_delay,
                            reason=str(last_error),
                            method=method,
                            url=url,
                        )
                    )
                    await asyncio.sleep(retry_delay)
                    continue
                self._logger.on_error(
                    ErrorInfo(error=last_error, method=method, url=url)
                )
                raise last_error from exc

            except (
                httpx.ConnectError,
                httpx.NetworkError,
                httpx.RemoteProtocolError,
            ) as exc:
                last_error = APIConnectionError(f"Connection error: {exc}", cause=exc)
                if attempt < self._max_retries:
                    retry_delay = _calculate_delay(attempt)
                    self._logger.on_retry(
                        RetryInfo(
                            attempt=attempt + 1,
                            max_retries=self._max_retries,
                            delay_seconds=retry_delay,
                            reason=str(last_error),
                            method=method,
                            url=url,
                        )
                    )
                    await asyncio.sleep(retry_delay)
                    continue
                self._logger.on_error(
                    ErrorInfo(error=last_error, method=method, url=url)
                )
                raise last_error from exc

        assert last_error is not None
        raise last_error

    async def aclose(self) -> None:
        if self._owns_client:
            await self._client.aclose()


# ── Shared helpers ───────────────────────────────────────────────────────


def _parse_remaining_req(response: httpx.Response) -> RemainingRequest | None:
    header = response.headers.get("Remaining-Req")
    if header:
        return RemainingRequest.from_header(header)
    return None



from __future__ import annotations

from functools import cached_property
from typing import Any

import httpx

from upbeat._auth import Credentials
from upbeat._config import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT, Timeout
from upbeat._constants import API_BASE_URL
from upbeat._http import AsyncTransport, SyncTransport
from upbeat._logger import Logger
from upbeat.api.accounts import AccountsAPI, AsyncAccountsAPI
from upbeat.api.deposits import AsyncDepositsAPI, DepositsAPI
from upbeat.api.markets import AsyncMarketsAPI, MarketsAPI
from upbeat.api.orders import AsyncOrdersAPI, OrdersAPI
from upbeat.api.quotation import AsyncQuotationAPI, QuotationAPI
from upbeat.api.travel_rule import AsyncTravelRuleAPI, TravelRuleAPI
from upbeat.api.withdrawals import AsyncWithdrawalsAPI, WithdrawalsAPI
from upbeat.shortcuts import AsyncShortcuts, SyncShortcuts
from upbeat.strategies import AsyncStrategies, SyncStrategies
from upbeat.ws._client import AsyncUpbeatWebSocket

# ── Upbeat (sync client) ────────────────────────────────────────────────


class Upbeat:
    _transport: SyncTransport
    _credentials: Credentials | None
    _owns_http_client: bool
    _closed: bool

    def __init__(
        self,
        *,
        access_key: str | None = None,
        secret_key: str | None = None,
        base_url: str = API_BASE_URL,
        timeout: Timeout = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        auto_throttle: bool = True,
        logger: Logger | None = None,
        http_client: httpx.Client | None = None,
        event_hooks: dict[str, list[Any]] | None = None,
    ) -> None:
        if (access_key is None) != (secret_key is None):
            raise ValueError(
                "Both access_key and secret_key must be provided together, or neither."
            )

        self._credentials = (
            Credentials(access_key=access_key, secret_key=secret_key)
            if access_key is not None and secret_key is not None
            else None
        )

        self._base_url = base_url
        self._timeout = timeout
        self._max_retries = max_retries
        self._auto_throttle = auto_throttle
        self._logger = logger
        self._owns_http_client = http_client is None
        self._closed = False

        self._transport = SyncTransport(
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            auto_throttle=auto_throttle,
            logger=logger,
            http_client=http_client,
            event_hooks=event_hooks,
        )

    # ── Resource properties ──────────────────────────────────────────

    @cached_property
    def quotation(self) -> QuotationAPI:
        return QuotationAPI(self._transport, self._credentials)

    @cached_property
    def accounts(self) -> AccountsAPI:
        return AccountsAPI(self._transport, self._credentials)

    @cached_property
    def orders(self) -> OrdersAPI:
        return OrdersAPI(self._transport, self._credentials)

    @cached_property
    def deposits(self) -> DepositsAPI:
        return DepositsAPI(self._transport, self._credentials)

    @cached_property
    def withdrawals(self) -> WithdrawalsAPI:
        return WithdrawalsAPI(self._transport, self._credentials)

    @cached_property
    def travel_rule(self) -> TravelRuleAPI:
        return TravelRuleAPI(self._transport, self._credentials)

    @cached_property
    def markets(self) -> MarketsAPI:
        return MarketsAPI(self._transport, self._credentials)

    @cached_property
    def strategies(self) -> SyncStrategies:
        return SyncStrategies(self)

    @cached_property
    def shortcuts(self) -> SyncShortcuts:
        return SyncShortcuts(self)

    # ── Lifecycle ────────────────────────────────────────────────────

    def close(self) -> None:
        if not self._closed and self._owns_http_client:
            self._transport.close()
        self._closed = True

    def __enter__(self) -> Upbeat:
        return self

    def __exit__(self, *_args: object) -> None:
        self.close()

    # ── with_options ─────────────────────────────────────────────────

    def with_options(
        self,
        *,
        max_retries: int | None = None,
        auto_throttle: bool | None = None,
        logger: Logger | None = None,
    ) -> Upbeat:
        new = Upbeat.__new__(Upbeat)
        new._credentials = self._credentials
        new._base_url = self._base_url
        new._timeout = self._timeout
        new._max_retries = max_retries if max_retries is not None else self._max_retries
        new._auto_throttle = (
            auto_throttle if auto_throttle is not None else self._auto_throttle
        )
        new._logger = logger if logger is not None else self._logger
        new._owns_http_client = False
        new._closed = False

        new._transport = SyncTransport(
            base_url=new._base_url,
            timeout=new._timeout,
            max_retries=new._max_retries,
            auto_throttle=new._auto_throttle,
            logger=new._logger,
            http_client=self._transport._client,
        )
        return new


# ── AsyncUpbeat (async client) ───────────────────────────────────────────


class AsyncUpbeat:
    _transport: AsyncTransport
    _credentials: Credentials | None
    _owns_http_client: bool
    _closed: bool

    def __init__(
        self,
        *,
        access_key: str | None = None,
        secret_key: str | None = None,
        base_url: str = API_BASE_URL,
        timeout: Timeout = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        auto_throttle: bool = True,
        logger: Logger | None = None,
        http_client: httpx.AsyncClient | None = None,
        event_hooks: dict[str, list[Any]] | None = None,
    ) -> None:
        if (access_key is None) != (secret_key is None):
            raise ValueError(
                "Both access_key and secret_key must be provided together, or neither."
            )

        self._credentials = (
            Credentials(access_key=access_key, secret_key=secret_key)
            if access_key is not None and secret_key is not None
            else None
        )

        self._base_url = base_url
        self._timeout = timeout
        self._max_retries = max_retries
        self._auto_throttle = auto_throttle
        self._logger = logger
        self._owns_http_client = http_client is None
        self._closed = False

        self._transport = AsyncTransport(
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            auto_throttle=auto_throttle,
            logger=logger,
            http_client=http_client,
            event_hooks=event_hooks,
        )

    # ── Resource properties ──────────────────────────────────────────

    @cached_property
    def quotation(self) -> AsyncQuotationAPI:
        return AsyncQuotationAPI(self._transport, self._credentials)

    @cached_property
    def accounts(self) -> AsyncAccountsAPI:
        return AsyncAccountsAPI(self._transport, self._credentials)

    @cached_property
    def orders(self) -> AsyncOrdersAPI:
        return AsyncOrdersAPI(self._transport, self._credentials)

    @cached_property
    def deposits(self) -> AsyncDepositsAPI:
        return AsyncDepositsAPI(self._transport, self._credentials)

    @cached_property
    def withdrawals(self) -> AsyncWithdrawalsAPI:
        return AsyncWithdrawalsAPI(self._transport, self._credentials)

    @cached_property
    def travel_rule(self) -> AsyncTravelRuleAPI:
        return AsyncTravelRuleAPI(self._transport, self._credentials)

    @cached_property
    def markets(self) -> AsyncMarketsAPI:
        return AsyncMarketsAPI(self._transport, self._credentials)

    @cached_property
    def ws(self) -> AsyncUpbeatWebSocket:
        return AsyncUpbeatWebSocket(self._credentials, self._logger)

    @cached_property
    def strategies(self) -> AsyncStrategies:
        return AsyncStrategies(self)

    @cached_property
    def shortcuts(self) -> AsyncShortcuts:
        return AsyncShortcuts(self)

    # ── Lifecycle ────────────────────────────────────────────────────

    async def aclose(self) -> None:
        if not self._closed and self._owns_http_client:
            await self._transport.aclose()
        self._closed = True

    async def __aenter__(self) -> AsyncUpbeat:
        return self

    async def __aexit__(self, *_args: object) -> None:
        await self.aclose()

    # ── with_options ─────────────────────────────────────────────────

    def with_options(
        self,
        *,
        max_retries: int | None = None,
        auto_throttle: bool | None = None,
        logger: Logger | None = None,
    ) -> AsyncUpbeat:
        new = AsyncUpbeat.__new__(AsyncUpbeat)
        new._credentials = self._credentials
        new._base_url = self._base_url
        new._timeout = self._timeout
        new._max_retries = max_retries if max_retries is not None else self._max_retries
        new._auto_throttle = (
            auto_throttle if auto_throttle is not None else self._auto_throttle
        )
        new._logger = logger if logger is not None else self._logger
        new._owns_http_client = False
        new._closed = False

        new._transport = AsyncTransport(
            base_url=new._base_url,
            timeout=new._timeout,
            max_retries=new._max_retries,
            auto_throttle=new._auto_throttle,
            logger=new._logger,
            http_client=self._transport._client,
        )
        return new



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Timeout:
    connect: float = 10.0
    read: float = 30.0


DEFAULT_TIMEOUT = Timeout()
DEFAULT_MAX_RETRIES = 3


@dataclass(frozen=True, slots=True)
class UpbeatConfig:
    timeout: Timeout = DEFAULT_TIMEOUT
    max_retries: int = DEFAULT_MAX_RETRIES
    auto_throttle: bool = True



from __future__ import annotations

import hashlib
import uuid
from urllib.parse import unquote, urlencode

import jwt


class Credentials:
    access_key: str
    secret_key: str

    def __init__(self, access_key: str, secret_key: str) -> None:
        if not access_key or not access_key.strip():
            raise ValueError("access_key must not be empty or blank")
        if not secret_key or not secret_key.strip():
            raise ValueError("secret_key must not be empty or blank")
        if " " in access_key or "\t" in access_key:
            raise ValueError("access_key must not contain whitespace")
        if " " in secret_key or "\t" in secret_key:
            raise ValueError("secret_key must not contain whitespace")
        self.access_key = access_key
        self.secret_key = secret_key

    def __repr__(self) -> str:
        if len(self.access_key) > 6:
            masked = f"{self.access_key[:3]}...{self.access_key[-3:]}"
        else:
            masked = self.access_key[:1] + "***"
        return f"Credentials(access_key='{masked}', secret_key='***')"


def build_query_string(params: dict[str, object]) -> str:
    if not params:
        return ""
    return unquote(urlencode(params, doseq=True))


def build_jwt(credentials: Credentials, query_string: str = "") -> str:
    payload: dict[str, str] = {
        "access_key": credentials.access_key,
        "nonce": str(uuid.uuid4()),
    }

    if query_string:
        query_hash = hashlib.sha512(query_string.encode("utf-8")).hexdigest()
        payload["query_hash"] = query_hash
        payload["query_hash_alg"] = "SHA512"

    return jwt.encode(payload, credentials.secret_key, algorithm="HS512")


def build_auth_header(credentials: Credentials, query_string: str = "") -> str:
    return f"Bearer {build_jwt(credentials, query_string)}"



from decimal import ROUND_DOWN, Decimal
from enum import StrEnum

API_BASE_URL = "https://api.upbit.com"
WS_PUBLIC_URL = "wss://api.upbit.com/websocket/v1"
WS_PRIVATE_URL = "wss://api.upbit.com/websocket/v1/private"

# ── KRW tick size table ──────────────────────────────────────────────────

_KRW_TICK_TABLE: list[tuple[Decimal, Decimal]] = [
    (Decimal("2_000_000"), Decimal("1000")),
    (Decimal("1_000_000"), Decimal("500")),
    (Decimal("500_000"), Decimal("100")),
    (Decimal("100_000"), Decimal("50")),
    (Decimal("10_000"), Decimal("10")),
    (Decimal("1_000"), Decimal("5")),
    (Decimal("100"), Decimal("1")),
    (Decimal("10"), Decimal("0.1")),
    (Decimal("1"), Decimal("0.01")),
]

_KRW_TICK_DEFAULT = Decimal("0.001")


def get_krw_tick_size(price: Decimal) -> Decimal:
    for threshold, tick_size in _KRW_TICK_TABLE:
        if price >= threshold:
            return tick_size
    return _KRW_TICK_DEFAULT


def round_to_tick(price: Decimal) -> Decimal:
    tick = get_krw_tick_size(price)
    return (price / tick).to_integral_value(rounding=ROUND_DOWN) * tick


# ── StrEnum constants ────────────────────────────────────────────────────


class OrderSide(StrEnum):
    BID = "bid"
    ASK = "ask"


class OrderType(StrEnum):
    LIMIT = "limit"
    PRICE = "price"
    MARKET = "market"
    BEST = "best"


class OrderState(StrEnum):
    WAIT = "wait"
    WATCH = "watch"
    DONE = "done"
    CANCEL = "cancel"


class TimeInForce(StrEnum):
    IOC = "ioc"
    FOK = "fok"
    POST_ONLY = "post_only"


class MarketState(StrEnum):
    PREVIEW = "preview"
    ACTIVE = "active"
    DELISTED = "delisted"


class SortOrder(StrEnum):
    ASC = "asc"
    DESC = "desc"



from __future__ import annotations

from upbeat.strategies._base import (
    AsyncClientProtocol,
    PortfolioItem,
    SyncClientProtocol,
)


def get_portfolio_value(client: SyncClientProtocol) -> float:
    accounts = client.accounts.list()

    krw_total = 0.0
    non_krw = []

    for account in accounts:
        balance = float(account.balance) + float(account.locked)
        if account.currency == "KRW":
            krw_total += balance
        elif balance > 0:
            non_krw.append(account)

    if non_krw:
        markets = ",".join(f"KRW-{a.currency}" for a in non_krw)
        tickers = client.quotation.get_tickers(markets)
        price_map = {t.market: t.trade_price for t in tickers}

        for account in non_krw:
            market = f"KRW-{account.currency}"
            balance = float(account.balance) + float(account.locked)
            price = price_map.get(market, 0.0)
            krw_total += balance * price

    return krw_total


def get_portfolio(client: SyncClientProtocol) -> list[PortfolioItem]:
    accounts = client.accounts.list()

    non_krw = []
    items: list[PortfolioItem] = []

    for account in accounts:
        balance = float(account.balance)
        locked = float(account.locked)
        if account.currency == "KRW":
            items.append(
                PortfolioItem(
                    currency="KRW",
                    balance=balance,
                    locked=locked,
                    current_price=1.0,
                    market="KRW",
                    estimated_value_krw=balance + locked,
                )
            )
        elif balance + locked > 0:
            non_krw.append(account)

    if non_krw:
        markets = ",".join(f"KRW-{a.currency}" for a in non_krw)
        tickers = client.quotation.get_tickers(markets)
        price_map = {t.market: t.trade_price for t in tickers}

        for account in non_krw:
            market = f"KRW-{account.currency}"
            balance = float(account.balance)
            locked = float(account.locked)
            price = price_map.get(market, 0.0)
            items.append(
                PortfolioItem(
                    currency=account.currency,
                    balance=balance,
                    locked=locked,
                    current_price=price,
                    market=market,
                    estimated_value_krw=(balance + locked) * price,
                )
            )

    return items


async def async_get_portfolio_value(client: AsyncClientProtocol) -> float:
    accounts = await client.accounts.list()

    krw_total = 0.0
    non_krw = []

    for account in accounts:
        balance = float(account.balance) + float(account.locked)
        if account.currency == "KRW":
            krw_total += balance
        elif balance > 0:
            non_krw.append(account)

    if non_krw:
        markets = ",".join(f"KRW-{a.currency}" for a in non_krw)
        tickers = await client.quotation.get_tickers(markets)
        price_map = {t.market: t.trade_price for t in tickers}

        for account in non_krw:
            market = f"KRW-{account.currency}"
            balance = float(account.balance) + float(account.locked)
            price = price_map.get(market, 0.0)
            krw_total += balance * price

    return krw_total


async def async_get_portfolio(client: AsyncClientProtocol) -> list[PortfolioItem]:
    accounts = await client.accounts.list()

    non_krw = []
    items: list[PortfolioItem] = []

    for account in accounts:
        balance = float(account.balance)
        locked = float(account.locked)
        if account.currency == "KRW":
            items.append(
                PortfolioItem(
                    currency="KRW",
                    balance=balance,
                    locked=locked,
                    current_price=1.0,
                    market="KRW",
                    estimated_value_krw=balance + locked,
                )
            )
        elif balance + locked > 0:
            non_krw.append(account)

    if non_krw:
        markets = ",".join(f"KRW-{a.currency}" for a in non_krw)
        tickers = await client.quotation.get_tickers(markets)
        price_map = {t.market: t.trade_price for t in tickers}

        for account in non_krw:
            market = f"KRW-{account.currency}"
            balance = float(account.balance)
            locked = float(account.locked)
            price = price_map.get(market, 0.0)
            items.append(
                PortfolioItem(
                    currency=account.currency,
                    balance=balance,
                    locked=locked,
                    current_price=price,
                    market=market,
                    estimated_value_krw=(balance + locked) * price,
                )
            )

    return items



from __future__ import annotations

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.market import TradingPair


class MarketsAPI(_SyncAPIResource):
    def get_trading_pairs(self, *, is_details: bool = False) -> list[TradingPair]:
        params = {"is_details": "true" if is_details else "false"}
        response = self._transport.request("GET", "/v1/market/all", params=params)
        return [TradingPair.model_validate(item) for item in response.data]


class AsyncMarketsAPI(_AsyncAPIResource):
    async def get_trading_pairs(self, *, is_details: bool = False) -> list[TradingPair]:
        params = {"is_details": "true" if is_details else "false"}
        response = await self._transport.request("GET", "/v1/market/all", params=params)
        return [TradingPair.model_validate(item) for item in response.data]



from __future__ import annotations

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.account import Account


class AccountsAPI(_SyncAPIResource):
    def list(self) -> list[Account]:
        response = self._transport.request(
            "GET", "/v1/accounts", credentials=self._credentials
        )
        return [Account.model_validate(item) for item in response.data]


class AsyncAccountsAPI(_AsyncAPIResource):
    async def list(self) -> list[Account]:
        response = await self._transport.request(
            "GET", "/v1/accounts", credentials=self._credentials
        )
        return [Account.model_validate(item) for item in response.data]



from __future__ import annotations

from upbeat._base import _AsyncAPIResource, _SyncAPIResource
from upbeat.types.travel_rule import TravelRuleVasp, TravelRuleVerification


class TravelRuleAPI(_SyncAPIResource):
    def list_vasps(self) -> list[TravelRuleVasp]:
        response = self._transport.request(
            "GET", "/v1/travel_rule/vasps", credentials=self._credentials
        )
        return [TravelRuleVasp.model_validate(item) for item in response.data]

    def verify_by_uuid(
        self,
        *,
        deposit_uuid: str,
        vasp_uuid: str,
    ) -> TravelRuleVerification:
        json_body = {"deposit_uuid": deposit_uuid, "vasp_uuid": vasp_uuid}
        response = self._transport.request(
            "POST",
            "/v1/travel_rule/deposit/uuid",
            json_body=json_body,
            credentials=self._credentials,
        )
        return TravelRuleVerification.model_validate(response.data)

    def verify_by_txid(
        self,
        *,
        vasp_uuid: str,
        txid: str,
        currency: str,
        net_type: str,
    ) -> TravelRuleVerification:
        json_body = {
            "vasp_uuid": vasp_uuid,
            "txid": txid,
            "currency": currency,
            "net_type": net_type,
        }
        response = self._transport.request(
            "POST",
            "/v1/travel_rule/deposit/txid",
            json_body=json_body,
            credentials=self._credentials,
        )
        return TravelRuleVerification.model_validate(response.data)


class AsyncTravelRuleAPI(_AsyncAPIResource):
    async def list_vasps(self) -> list[TravelRuleVasp]:
        response = await self._transport.request(
            "GET", "/v1/travel_rule/vasps", credentials=self._credentials
        )
        return [TravelRuleVasp.model_validate(item) for item in response.data]

    async def verify_by_uuid(
        self,
        *,
        deposit_uuid: str,
        vasp_uuid: str,
    ) -> TravelRuleVerification:
        json_body = {"deposit_uuid": deposit_uuid, "vasp_uuid": vasp_uuid}
        response = await self._transport.request(
            "POST",
            "/v1/travel_rule/deposit/uuid",
            json_body=json_body,
            credentials=self._credentials,
        )
        return TravelRuleVerification.model_validate(response.data)

    async def verify_by_txid(
        self,
        *,
        vasp_uuid: str,
        txid: str,
        currency: str,
        net_type: str,
    ) -> TravelRuleVerification:
        json_body = {
            "vasp_uuid": vasp_uuid,
            "txid": txid,
            "currency": currency,
            "net_type": net_type,
        }
        response = await self._transport.request(
            "POST",
            "/v1/travel_rule/deposit/txid",
            json_body=json_body,
            credentials=self._credentials,
        )
        return TravelRuleVerification.model_validate(response.data)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class MarketCaution(BaseModel):
    model_config = ConfigDict(frozen=True)

    PRICE_FLUCTUATIONS: bool = False
    TRADING_VOLUME_SOARING: bool = False
    DEPOSIT_AMOUNT_SOARING: bool = False
    GLOBAL_PRICE_DIFFERENCES: bool = False
    CONCENTRATION_OF_SMALL_ACCOUNTS: bool = False


class MarketEvent(BaseModel):
    model_config = ConfigDict(frozen=True)

    warning: bool
    caution: MarketCaution


class TradingPair(BaseModel):
    model_config = ConfigDict(frozen=True)

    market: str
    korean_name: str
    english_name: str
    market_event: MarketEvent | None = None



from __future__ import annotations

from upbeat.types.account import Account


def find_account(accounts: list[Account], currency: str) -> Account | None:
    currency_upper = currency.upper()
    for account in accounts:
        if account.currency == currency_upper:
            return account
    return None
